import os
import logging

from mysql.connector.cursor import MySQLCursorBuffered

from yangke.core import runCMD, enable_docker, status_of_container, str_in_list, runAsAdmin
from yangke.base import get_args, get_settings
from yangke.common.config import logger
import pymysql
from pymysql.connections import Connection, Cursor
from sqlalchemy import create_engine, MetaData, Table, NVARCHAR, Float, Date, Integer, Column, CHAR, DateTime, String
from sqlalchemy.engine.base import Engine
from sqlalchemy.orm import sessionmaker, scoped_session
import traceback
import warnings
import json
import re
import datetime
import pandas as pd

try:
    # noinspection all
    from DBUtils.PooledDB import PooledDB  # DBUtils 1.*版本写法
except ModuleNotFoundError:
    # noinspection all
    from dbutils.pooled_db import PooledDB  # DBUtils 2.*版本写法

# encoding = 'utf-8'
pool: PooledDB = None  # 连接池

dtype_dict = {
    'str': NVARCHAR(length=255),
    'float': Float(),
    'int': Integer(),
    'float64': Float(),
    'date': Date(),
}

# 最后一次调用该模块的方法时使用的engine，单线程可以不传入engine，默认使用上一次的
# 这里之所以不保存cursor，是因为pymysql对cursor不会做连接存活处理，经常发生cursor自动断开导致操作报错的情况。
# 报错类型为：pymysql.err.InterfaceError: (0, '')
last_engine: Engine = None


def __start_mysql__(name_of_container, volume, user, password, mysqlPort):
    def print_lines(lines):
        for line in lines:
            print(line)

    # =========================确保指定名称的容器没有运行，如果在运行则跳过启动操作=================================
    status = status_of_container(name_of_container)
    if status is None:
        pass
    elif "up" in status.lower():
        # 容器不存在，则跳过if代码块
        logging.debug("容器{}正在运行！".format(name_of_container))
        return True
    else:
        # 如果存在同名容器，且容器没有运行，则删除该容器
        runCMD("docker container rm --force " + name_of_container)

    # 该条docker命令中 -p 3306:3306最好显式指出，否则JDBC可能连接不上；而adminer无论是否指出端口都可以连接上。
    # MYSQL_USER、MYSQL_PASSWORD和MYSQL_DATABASE也是为了spring使用jdbc连接建立的数据库
    # docker run --rm --name some-mysql -p 3306:3306 -e MYSQL_ROOT_PASSWORD=111111 -e MYSQL_USER=springuser -e MYSQL_PASSWORD=ThePassword -e MYSQL_DATABASE=test -d --privileged=true mysql:latest
    # '-e MYSQL_USER=yangke -e MYSQL_PASSWORD=111111 -e MYSQL_DATABASE=test ' \
    if volume is not None:
        command = 'docker run --rm --name some-mysql -p {}:3306 -e MYSQL_ROOT_PASSWORD={} ' \
                  '-d -v "{}:/var/lib/mysql/" --privileged=true mysql:latest'.format(mysqlPort, password, volume)
    else:
        command = 'docker run --rm --name some-mysql -p {}:3306 -e MYSQL_ROOT_PASSWORD={} ' \
                  '-d --privileged=true mysql:latest'.format(mysqlPort, password)
    logging.debug("执行命令：" + command)
    logging.info('挂载mysql数据卷路径为: "{}"'.format(volume))
    result, err = runCMD(command)
    if str_in_list('mkdir /host_mnt/d: file exists', err):
        logging.info(err)
        print("Docker服务异常，导致docker的本地卷挂载不上请重启Dockers Desktop后再试!")
        return False
    elif str_in_list("doesn't exist and is not known to Docker", err):
        logging.info(err)
        print("挂载的卷不存在，请检查卷路径！")
        return False
    elif len(err) > 0:
        logging.info(err)
        return False
    else:
        logging.debug("MySQL服务已经启动，信息如下：")
        title, _ = runCMD('docker container ps | findstr "container"')
        print_lines(title)
        results, _ = runCMD('docker container ps | findstr "some-mysql"')
        print_lines(results)
        return True


def __start_adminer__(db_name, port, ui="nette"):
    """
    启动adminer服务，连接到指定的mysql数据库
    :param db_name: 运行mysql数据库服务端的容器名称
    :param port: 设置adminer映射到本地的端口号
    :param ui: 控制了adminer的界面风格，可选的参数参见https://www.adminer.org/
    :return:
    """
    # 这里列出几个ui风格
    # ui = "mancave"  # mancave-hever, pepa-linha-dark, galkaev, price, mancave
    # ===================== 确保没有 some-adminer 容器在运行==========================
    adminer_name = "some-adminer"
    status = status_of_container(adminer_name)
    if status is not None:
        runCMD('docker container rm --force {}'.format(adminer_name))
        logging.debug("发现已经创建的adminer服务，重新创建adminer服务！")
    # ========================== 启动 some-adminer 容器===============================
    status = status_of_container(db_name)
    if "up" in status.lower():
        # 如果需要设置adminer的界面，参考https://www.adminer.org/en/
        # 如果需要使用galkaev的界面，设置如下（但测试报错）
        # docker run --name {} --link {}:db -p {}:8080 -e ADMINER_DESIGN="galkaev" adminer
        command = 'docker run --name {} --link {}:db -e ADMINER_DESIGN="{}" -p {}:8080 -d adminer' \
            .format(adminer_name, db_name, ui, port)
        logging.debug('执行docker命令：' + command)
        result, err = runCMD(command)
        # 验证adminer是否正常启动
        import time
        time.sleep(1)
        status = status_of_container(adminer_name)
        if "up" in status.lower():
            print("您现在可以通过浏览器访问 http://localhost:{} 连接到mysql数据库".format(port))
        else:  # 说明启动失败，一般情况下，这时候adminer容器的状态是Created；重启一次adminer拿到失败的信息并输出
            logging.debug(result)
            logging.debug(err)
            if str_in_list("Error starting userland proxy: /forwards/expose/port returned unexpected status: 500", err):
                logging.debug('设置的adminer端口{}不可用，自动尝试端口{}...'.format(port, port + 1))
                __start_adminer__(db_name, port + 1, ui)
            else:
                print("Adminer启动失败，请检查配置！")
                return False
    else:
        logging.warning("Try to start Adminer, but no mysql container found...")
        return False
    return True


def connect_available(user=None, passwd=None, password=None, host=None, port=None,
                      db=None, charset="utf8mb4",
                      return_pool=False, config=None):
    """
    判断指定的参数是否可以连接到mysql服务器，参数含义参考pymysql.connect()方法。
    如果只需判断mysql服务是否可用，而不用判断用户名密码等连接参数是否正确，请使用mysql_available()方法

    :passwd=password,两个参数含义相同，是别名
    :return:
    """

    port = int(port)
    passwd = passwd or password
    result = connect_mysql(user, passwd, host, port,
                           db, charset=charset,
                           return_type="pool", config=config)
    if result is None:
        return False
    else:
        result.close()
        return True


def mysql_available(user=None, passwd=None, host=None, port=None, db=None, create_if_not_exist=True, charset=None,
                    config=None) -> bool:
    """
    测试mysql服务是否可用，不带参数，仅代表本地是否有mysql服务
    如果测试连接是否可用，请使用connect_available()方法

    :config:
    :return:
    """

    # 当带参数判断时，需要借助pymysql，因为如果mysql -u root是交互式的，runCMD会阻塞
    if user is not None and passwd is not None:
        cursor = connect_mysql(user=user, passwd=passwd, host=host, port=port, db=db,
                               create_if_not_exist=create_if_not_exist,
                               charset=charset, config=config)
        if cursor is None:
            return False
        else:
            return True
    # mysql -h 127.0.0.1 -P 3306 -u root -p
    _, err = runCMD("mysql", charset='gbk', wait_for_result=True)
    # 测试mysql8.0.18，运行'mysql'命令，结果如下：
    # 当mysql可用时，提示：ERROR 1045 (28000): Access denied for user 'ODBC'@'localhost' (using password: NO)
    # 当mysql不可用时，提示：ERROR 2003 (HY000): Can't connect to MySQL server on 'localhost' (10061)
    # windows返回ERROR 1045
    # ubuntu远程访问mysql返回ERROR 2002 (HY000): Can't connect to local MySQL server through
    # socket '/var/run/mysqld/mysqld.sock' (2)
    if str_in_list('ERROR 1045', err) or 'ERROR 1045' in err or 'ERROR 2002' in err:
        return True
    else:  # if strInList("ERROR 2003", err) or 'ERROR 2003' in err or 'mysql: not found' in err:
        return False


def start_mysql_service():
    """
    开启本地mysql服务，如果已启动，则不操作

    :return:
    """
    # args = getArgs('type', 'service_name', 'volume', 'adminer', 'user', 'password', 'mysql_port')

    service_type = 'service'  # 默认启动服务
    service_name = 'mysql'  # 默认启动的服务名
    configs = get_settings()
    if configs.get('mysql') is None:
        pass
    else:
        if configs.get('mysql').get('service') is not None:
            service_type = 'service'
        elif configs.get('mysql').get('docker') is not None:
            service_type = 'docker'

    if service_type == 'service':

        if configs.get('mysql') is not None:
            serviceDict = dict(configs.get('mysql').get('service'))
            service_name = serviceDict.get('service_name') or 'mysql'
        else:
            pass
        # 测试mysql是否可用
        if not mysql_available():
            from threading import Thread
            import platform
            if str(platform.platform()).startswith("Windows"):
                cmd = 'net start {}'.format(service_name)
                t1 = Thread(target=runAsAdmin, args=(cmd,))
            elif str(platform.platform().startswith("Linux")):
                cmd = 'service {} start'.format(service_name)
                t1 = Thread(target=runCMD, args=(cmd,))
            # 这里开个新线程启动mysql，因为当用户拒绝以管理员身份运行时，runAsAdmin命令会卡住

            t1.setDaemon(True)
            t1.start()
            t1.join(20)
            # runAsAdmin('net start {}'.format(service_name))

        if mysql_available():
            return True
        else:
            return False
    elif service_type == 'docker':
        docker_dict = dict(configs.get('mysql').get('docker'))
        user = docker_dict.get('user')
        password = docker_dict.get('passwd')
        volume = docker_dict.get('volume')
        adminer = docker_dict.get('adminer') or None
        mysql_port = docker_dict.get('port') or 3306
        adminer_port = docker_dict.get('adminerPort') or 8080
        container_name = docker_dict.get('containerName') or "some-mysql"
        success = enable_docker()
        assert success, "当前配置使用docker - mysql，但docker服务不可用"
        # path = os.getcwd() # 这个是调用者所在的路径
        success = __start_mysql__(container_name, volume=volume, user=user, password=password, mysqlPort=mysql_port)
        assert success, "当前配置使用docker - mysql，mysql容器启动失败"
        if adminer:
            success = __start_adminer__(container_name, adminer_port)
            if not success:
                return False
        return True


def connect_mysql(user="root", passwd="111111", host="localhost", port=3306,
                  db=None, create_if_not_exist=True, charset="utf8mb4",
                  return_type="cursor", config=None, test_mode=False, use_pymysql=True):
    """
    尝试连接mysql，如果连接成功，则返回return_type指定的对象；如果不传入参数，则从settings文件读取配置信息，
    如果settings文件中存在多个mysql配置，以最后一个为准。
    如果配置文件有配置参数且传入参数，则以配置参数为准。优先级配置文件>制定参数>config参数
    如果使用config连接，示例如下
    mysql_config = {
    'host': 'localhost',
    'port': 3306,
    'user': 'root',
    'passwd': '111111',
    'charset': 'utf8mb4'
    }
    conn, cursor=connect_mysql(config=mysql_config)  # 这里获得了一个没有指定数据库的cursor
    # 给config参数指定数据库，再次连接
    mysql_config['db'] = db_name
    conn = pymysql.connect(**mysql_config)  # 这里获得另一个指定了数据库的连接
    conn, cursor = connect_mysql(config=mysql_config)  # 这里获得连接了数据库的cursor

    当使用连接池时，示例如下：
    pool = connect_mysql(db='face', return_pool=True)  # 不指定用户名密码则connect_mysql方法从配置文件读取配置参数
    conn = pool.connection()
    cursor = conn.cursor()

    :param use_pymysql: 使用pymysql连接数据库，也可以使用mysqlconnector，目前二者使用engine连接都有一些问题，pymysql有警告但不影响使用
    :param user:
    :param passwd:
    :param host:
    :param port:
    :param db: 需要连接的db名，如果指定的db不存在，该方法会自动创建
    :param create_if_not_exist: 当db不存在时，是否创建
    :param charset:
    :param return_type: “cursor"/"pool"/"engine"，返回类型，分别对应游标，连接池和sqlalchemy的引擎
    :param config:
    :param test_mode: 如果是测试连接是否可用，则不输出错误信息，如果是需要连接，则失败时会输出错误信息
    :return: 根据return_cursor的取值，返回不同数据
    """
    # 依次查询settings.py、config入参、其余指定项入参，存储到config_mysql中
    global pool
    configs = get_settings().get('mysql') or {}
    if configs is not None:  # 读取配置文件的配置
        mysql_configs = configs.get('service') or configs.get('docker') or configs  # 连接时无论那种服务类型，连接方式都是一致的，无需区分
        host = mysql_configs.get('host') or host  # 优先使用传入的参数，没有传入参数，则使用配置文件桉树
        port = mysql_configs.get('port') or port
        user = mysql_configs.get('user') or user
        passwd = mysql_configs.get('passwd') or passwd
        passwd = str(passwd) if passwd is not None else None  # 如果passwd不为None，则将passwd转换为str类型，防止纯数字密度读入为整形
        db = mysql_configs.get('db') or db
        charset = mysql_configs.get('charset') or charset or "utf8mb4"

    if config is not None:  # 如果传入参数不为空，则以传入参数为准
        host = host or config.get('host')
        port = port or config.get('port')
        user = user or config.get('user')
        passwd = passwd or config.get('passwd')
        charset = charset or config.get('charset')
        db = db or config.get('db')
    config_mysql = {'host': host,
                    'port': int(port),
                    'user': user,
                    'passwd': passwd,
                    'charset': charset,
                    'db': db}

    # 将config_mysql中的参数存到短变量名中，便于引用
    port = config_mysql['port']

    if return_type != "engine":
        # 建立mysql数据库连接
        try:
            if return_type == "pool":
                global pool
                pool = PooledDB(pymysql, 5, **config_mysql)  # 可能报错，现在没有考虑报错情况
                _update_cursor(pool.connection().cursor())
                return pool
            elif return_type == "cursor":
                conn = pymysql.connect(**config_mysql)
        except TypeError as e:
            if e.args[0] == "object supporting the buffer API required":  # 说明passwd是整数，目前在入口处已经转换类型
                logger.warning("输入的密码是整形数据，尝试转换为字符串后重新连接...")
        except pymysql.err.InternalError as e:
            if e.args[0] == 1049:  # 说明时不存在指定的db文件
                if create_if_not_exist:
                    config_mysql.pop('db')
                    conn = pymysql.connect(**config_mysql)
                    create_db(conn.cursor(), db_name=db)
                    return connect_mysql(user=user, passwd=passwd, host=host, port=port, db=db, charset=charset,
                                         return_type=return_type)
                else:
                    logger.error("需要连接的mysql数据库db文件不存在：{}".format(db))
                    return None
        except pymysql.err.OperationalError as e:  # 账号或密码错误
            if e.args[0] == 1045:  # (1045, "Access denied for user 'root'@'localhost' (using password: YES)")
                if not test_mode:
                    logger.error("mysql连接被拒绝，请检查账号密码，错误信息：{}".format(e))
            elif e.args[0] == 1044:  # (1044, "Access denied for user 'test'@'localhost' to database 'face'")
                if not test_mode:
                    logger.error("mysql连接被拒绝，连接 {} 的权限不足".format(e))
            elif e.args[
                0] == 1142:  # (1142, "CREATE command denied to user 'test'@'localhost' for table 'persongroup'")
                if not test_mode:
                    logger.error("mysql用户操作 {} 的权限不足".format(e))
            elif e.args[0] == 1049:  # 说明时不存在指定的db文件
                if create_if_not_exist:
                    config_mysql.pop('db')
                    conn = pymysql.connect(**config_mysql)
                    create_db(conn.cursor(), db_name=db)
                    return connect_mysql(user=user, passwd=passwd, host=host, port=port, db=db, charset=charset,
                                         return_type=return_type)
                else:
                    logger.error("需要连接的mysql数据库db文件不存在：{}".format(db))
                    return None
            return None
        # 创建游标
        cursor = conn.cursor()
        _update_cursor(cursor)
        return cursor
    else:
        if use_pymysql:
            eng = "pymysql"  # 有警告
        else:
            eng = "mysqlconnector"  # mysql8版本会报授权验证插件错误
        engine = create_engine(f'mysql+{eng}://{user}:{passwd}@{host}:{port}')
        if db is not None:
            if not exist_db(db_name=db, cursor=engine):  # 如果指定的db不存在，则创建
                create_db(db)
            engine = create_engine(f'mysql+{eng}://{user}:{passwd}@{host}:{port}/{db}')
        _update_cursor(engine)
        return engine


def exec_sql_script(sql_script=None, para=None, close=False, cursor=None):
    """
    执行指定的mysql语句

    Example:
        如果要执行的sql语句需要传入变量，使用如下方法：

        script = "insert into account(company, user, password, fund, stocks) VALUES (?,?,?,?,?)"
        para = (company, user, password, fund, stocks,)
        exec_sql_script(cursor, script, para)


    :param cursor:
    :param sql_script:
    :param para:
    :param commit: 是否需要提交，查询操作不需要提交，更新操作需要提交
    :param close: 操作完成后，是否需要关闭连接
    :return: 执行结果

    """
    cursor = _update_cursor(cursor)
    if isinstance(cursor, str):
        logger.error(
            f"当前传入的mysql语句执行游标为 cursor = {cursor}，该错误一般是因为漏掉了应该传入的游标参数")
        return None
    elif isinstance(cursor, Engine):
        cursor: Engine = cursor
        session_ = sessionmaker(bind=cursor)
        session = scoped_session(session_)
        res = session.execute(sql_script)
        session.commit()
        # res = cursor.execute(sql_script)  # engine只能查，修改的话无法commit
        if res.returns_rows:  # 如果执行sql语句有返回结果
            tmp = res.fetchall()
        else:
            tmp = []
        session.remove()
    elif isinstance(cursor, Cursor):
        if para is not None and not isinstance(para, tuple):
            para = (para,)
        cursor.execute(sql_script, para)
        tmp = cursor.fetchall()
        if close:
            cursor.close()
            cursor.connection.close()
    elif isinstance(cursor, MySQLCursorBuffered):
        raise ValueError("暂不支持mysqlconnector的游标")
    return tmp


def _update_cursor(engine):
    """
    记录最后一次使用的游标。如果传入非空，则记录传入参数对应的cursor后返回原参数。如果传入为空，则返回记录的最后一次保存的cursor

    :param engine:
    :return:
    """
    global last_engine
    if engine is None or engine == last_engine:
        return last_engine
    else:
        if isinstance(engine, Cursor):
            last_engine = cursor2engine(engine)
        elif isinstance(engine, Engine):
            last_engine = engine
            return engine
        elif isinstance(engine, MySQLCursorBuffered):
            raise ValueError("暂不支持mysql-connector的游标")


def exist_db(db_name: str = None, cursor: Cursor or Engine = None):
    """判断数据库是否存在，需要确保cursor对应的mysql账号有权限范围对应的db，否则返回False"""
    cursor = _update_cursor(cursor)

    sql = "show databases like '{}'".format(db_name)
    if isinstance(cursor, Cursor):
        result = exec_sql_script(cursor=cursor, sql_script=sql)
        return False if len(result) == 0 else True
    else:
        result = cursor.execute(sql)
        return False if result.rowcount == 0 else True


def engine2cursor(engine: Engine):
    """
    从sqlalchemy的engine拿到对应的cursor，同一个engine多次调用该方法返回不同的cursor

    :param engine:
    :return:
    """
    _update_cursor(engine)  # 更新模块变量last_engine，记录对应的engine
    cursor = engine.pool.connect().cursor()  # 获取engine的cursor
    return cursor


def cursor2engine(cursor: Cursor = None):
    """
    将pymysql的cursor获取sqlalchemy的engine，同一个cursor多次调用该方法返回不同的engine

    :param cursor:
    :return:
    """
    # 这里不能添加_update_cursor()方法，以免行程递归死锁调用
    conn = cursor.connection
    encoding = conn.encoding
    user = conn.user.decode(encoding)
    passwd = conn.password.decode(encoding)
    host = conn.host
    port = conn.port
    db = conn.db.decode(encoding) if conn.db is not None else None
    engine = connect_mysql(user, passwd, host, port, db, return_type="engine")

    return engine


def create_db(db_name: str = None, cursor: Cursor or Engine = None):
    """
    创建数据库文件，确保有root权限
    """
    cursor = _update_cursor(cursor)  # 更新模块变量last_cursor
    sql = "create database `{}`".format(db_name)
    if not exist_db(db_name):
        if isinstance(cursor, Engine):
            _ = cursor.execute(sql)
        else:
            _ = exec_sql_script(sql)
    return True


def has_table(table_name: str = None, cursor: Cursor or Engine = None):
    """
    判断给定代码的股票的数据库表是否存在。确保mysql用户有操作权限

    GRANT ALL on face.* to 'test'@'localhost';

    :param cursor: 游标或engine
    :param table_name: 表名
    :return: bool
    """
    cursor = _update_cursor(cursor)  # 更新模块变量last_cursor
    if isinstance(cursor, Cursor):
        sql = "show tables like '{}'".format(table_name)
        result = exec_sql_script(sql, cursor=cursor)
        return False if len(result) == 0 else True
    else:
        cursor: Engine = cursor
        return cursor.has_table(table_name)


def create_table(table_name: str = None, columns: dict = None, primary=[0],
                 foreign=None,
                 metadata=None, cursor: Cursor or Engine = None) -> bool:
    """
    创建mysql表
    例1：
    from sqlalchemy import Column, Integer, String, Date, CHAR, Table, MetaData
    holiday_table = MetaData()  # 没有关系的表放在不同的metadata中
    Table('holiday', holiday_table,
          Column('calendarDate', Date, primary_key=True),
          Column('isOpen', CHAR, nullable=False))
    engine = sqlalchemy.create_engine("mysql+pymysql://root:111111@localhost:3306/stocks")
    create_table(cursor=engine, metadata=holiday_table)
    如果使用sqlalchemy的metadata创建表，第一个参数必须传入Engine类型对象

    例2：
    create_table(cursor=cursor, table_name="item", columns={"id": "int", "name": "varchar(20)", "price": "float"})
    其中columns的取值参见mysql建表语句中的数据类型说明

    :param metadata: 如果cursor是sqlalchemy的engine，则table为对应的sqlalchemy的metadata对象
    :param cursor: 连接了数据库的游标或engine
    :param table_name: 表名
    :param columns: 列名和数据类型的字典，如{"id": "int", "name": "varchar(20)", "price": "float"}，日期最好用varchar，使用python转换
    :param primary: 主键，默认第一个，多个主键通过[0,1 ]等列索引传入
    :param foreign: 外键引用信息，例如foreign="FOREIGN KEY (person_id) REFERENCES person(id) ON DELETE CASCADE ON UPDATE CASCADE"
    :return: bool
    """
    cursor = _update_cursor(cursor)  # 更新模块变量last_cursor
    if isinstance(cursor, Engine):
        if metadata is not None:
            metadata.metadata.create_all(cursor)
            return True
        else:
            engine: Engine = cursor
            cursor = engine2cursor(engine)
    if isinstance(primary, int):
        primary = [primary]
    primaryKey = []
    i = 0

    temp = ""
    for key, data_type in columns.items():
        temp = temp + "`{}` {},".format(key, data_type)
        if i in primary:
            primaryKey.append("`{}`".format(key))
        i = i + 1
    primary_str = ",".join(primaryKey)
    if foreign is not None:
        foreign = ", {}".format(foreign)
    else:
        foreign = ""
    sql = "create table if not exists `{}` ({} primary key ({}) {});".format(table_name, temp, primary_str, foreign)
    exec_sql_script(sql_script=sql)
    return True


def delete_in_table(table_name: str = None, col_name: str = None, value: str = None, return_result=False, cursor=None,
                    **kwargs):
    """
    调用之前确认要删除的表和数据行存在

    :param return_result:
    :param cursor:
    :param table_name:
    :param col_name:
    :param value:
    :return:
    """
    cursor = _update_cursor(cursor)  # 更新模块变量last_cursor
    if len(kwargs) == 0:
        if col_name is None or value is None:
            return False
        sql = "delete from `{}` where `{}`='{}';".format(table_name, col_name, value)
    else:
        k_v_pairs = []
        for k, v in kwargs.items():
            k_v_pairs.append("`{}`='{}'".format(k, v))
        conditions = " and ".join(k_v_pairs)
        if conditions.strip() is None:
            return False
        sql = "delete from `{}` where {};".format(table_name, conditions)
    fetchall = exec_sql_script(cursor=cursor, sql_script=sql)
    if return_result:
        return True, fetchall
    else:
        return True


def exists_in_table(table_name: str = None, col_name: str = None, value: str = None,
                    condition_dict: dict = None,
                    return_result: bool = False, cursor=None):
    """
    表tableName中是否存在列col_name的值位value的行

    :param cursor: 连接了db文件的游标
    :param table_name: 表名
    :param col_name: 列名
    :param value: 列的值
    :param condition_dict: 查询的键值对字典值，优先于col_name和value传入的值，即会覆盖col_name和value传入的值
    :param return_result: 是否需要返回查找到的数据行，如果为真，则返回所有符合查找条件的数据行
    :return:
    """
    cursor = _update_cursor(cursor)  # 更新模块变量last_cursor
    if has_table(table_name, cursor):
        pass
    else:
        return False

    if condition_dict is not None and len(condition_dict) > 0:
        condition_str = []
        for k, v in condition_dict.items():
            condition_str.append("`{}`='{}'".format(k, v))
        condition_str = " and ".join(condition_str)
        col_name = k  # 在不需要返回结果是，随便查找一列
    else:
        condition_str = "`{}`='{}'".format(col_name, value)

    if return_result:
        sql = "(select * from `{}` where {});".format(table_name, condition_str)
    else:
        sql = "(select `{}` from `{}` where {});".format(col_name, table_name, condition_str)

    result = exec_sql_script(cursor=cursor, sql_script=sql)
    if return_result:
        return len(result) != 0, result
    else:
        return len(result) != 0


def insert_dataframe_to_mysql(table_name: str = None,
                              dataframe: pd.Series or pd.DataFrame = None,
                              ignore=False,
                              replace=False, filter_warning=None, cursor: Cursor or Engine = None):
    """
    因为pandas.DataFrame类的to_sql()方法存在一下问题：
    1.to_sql()方法自动创建的表的数据类型难以设置，而外键和PrimaryKey无法设置
    2.to_sql()方法插入dataframe数据时，如果表存在，则只能完全代替或完全追加，但当dataframe中有部分数据和数据库中
    相同时，无法只添加不同的数据，忽略相同的数据。

    该方法可以解决以上问题，尤其是第2个问题。

    将dataframe插入到mysql数据库中，如果dataframe中的某列列名不存在与mysql表中，则该列被忽略。如果mysql中的某列不存在于dataframe中，且
    mysql表定义时设置该列可以为空，则可以正常插入。

    :param cursor:
    :param table_name:
    :param dataframe: 插入的数据，可以是pd.DataFrame或pd.Series
    :param ignore: 当插入数据重复时，是否忽略，ignore和replace不能同时为True
    :param replace: 当插入数据重复时，是否替换，ignore和replace不能同时为True
    :param filter_warning: 过滤警告信息，[1062, 1265, 1366]，分别对应["Duplicate", "Data truncated", "Incorrect integer value"]
    :return:
    """
    cursor = _update_cursor(cursor)  # 更新模块变量last_cursor
    if isinstance(cursor, Engine):
        cursor: Cursor = engine2cursor(cursor)
    assert (ignore and replace) is False, "ignore和replace不能同时为True"
    ignore_ = "ignore" if ignore else ""
    if isinstance(dataframe, pd.DataFrame):
        row_num, col_num = dataframe.shape
        columns_df = set(dataframe.columns)
        columns_sql = set(get_column_name_of_table(table_name))
        columns_common = list(columns_sql & columns_df)
        columns_common.sort()
        value_list = []
        for i in range(row_num):
            temp_value = str(tuple([dataframe[name][i] for name in columns_common]))
            value_list.append(temp_value)
        values_str = ",".join(value_list)
    elif isinstance(dataframe, pd.Series):
        n = dataframe.size
        columns_df = set(dataframe.index)
        columns_sql = set(get_column_name_of_table(table_name))
        columns_common = list(columns_sql & columns_df)
        columns_common.sort()
        values_str = str(tuple([dataframe[name] for name in columns_common]))
    else:
        logger.warning(f"传入的数据类型{type(dataframe)}错误，本方法只支持插入DataFrame或Series类型")
        return None

    columns_str = ",".join([f'`{col}`' for col in columns_common])
    if replace:
        sql = f"REPLACE INTO `{table_name}` ({columns_str}) VALUES {values_str};"
    else:
        sql = "INSERT {} INTO `{}` ({}) VALUES {};".format(ignore_, table_name, columns_str, values_str)

    if filter_warning is None:
        exec_sql_script(sql_script=sql)
    elif isinstance(filter_warning, list):
        try:
            warnings.filterwarnings('error')
            exec_sql_script(cursor=cursor, sql_script=sql)
        except Warning as e:
            if e.args[0] in filter_warning:
                pass
            else:
                traceback.print_exc()


def insert_item_to_mysql(table_name: str = None, values: list = None,
                         col_names: list = None, ignore=False,
                         replace=False, filter_warning=None, cursor: Cursor or Engine = None):
    """
    向mysql数据库中插入数据，这里传入的列名不要用反引号括起来，增

    :param cursor: 连接了db文件的游标
    :param table_name: 表名
    :param values:
    :param col_names: 列名，如果是插入带有auto_increment属性的表数据，则必须指定列名，否则就需要指定auto_increment属性的字段的值
    :param ignore: 当插入数据重复时，是否忽略
    :param replace: 当插入数据重复时，是否替换，ignore和replace不能同时为True
    :param filter_warning: 过滤警告信息，[1062, 1265, 1366]，分别对应["Duplicate", "Data truncated", "Incorrect integer value"]
    """
    # 注意需要将python中的时间日期类型的数据转换为str插入
    import datetime
    cursor = _update_cursor(cursor)  # 更新模块变量last_cursor
    if isinstance(cursor, Cursor):  # 都转换为engine执行
        cursor = cursor2engine(cursor)
    # for index, value in enumerate(values):
        # if type(value) in [datetime.date, datetime.datetime, datetime.time]:
        #     values[index] = str(value)
    columns = ""
    if col_names is not None:
        columns = ",".join(
            ["`{}`".format(col) for col in col_names])  # linux下可能要用单引号括起来列名 ",".join("'{}'".format(col_names))
        columns = f"({columns})"
    # 当列名有特殊符号时，不能用?作为占位符进行数据插入
    values = ",".join(["'{}'".format(value) for value in values])
    assert (ignore and replace) is False, "ignore和replace不能同时为True"
    ignore_ = "ignore" if ignore else ""
    if replace:
        sql = "REPLACE INTO `{}` {} VALUES ({});".format(table_name, columns, values)
    else:
        sql = "INSERT {} INTO `{}` {} VALUES ({});".format(ignore_, table_name, columns, values)
    if filter_warning is None:
        exec_sql_script(cursor=cursor, sql_script=sql)
    elif isinstance(filter_warning, list):
        try:
            warnings.filterwarnings('error')
            exec_sql_script(cursor=cursor, sql_script=sql)
        except Warning as e:
            if e.args[0] in filter_warning:
                pass
            else:
                traceback.print_exc()


def update_in_table(table_name=None, condition_dict: dict = None, update_dict: dict = None, cursor=None) -> None:
    """
    更改数据库的值
    :param cursor:
    :param table_name:
    :param condition_dict:
    :param update_dict:
    :return:
    """
    _update_cursor(cursor)  # 更新模块变量last_cursor
    condition_str = []
    update_str = []
    for k, v in condition_dict.items():
        condition_str.append("`{}`='{}'".format(k, v))
    condition_str = " and ".join(condition_str)
    for k, v in update_dict.items():
        update_str.append("`{}`='{}'".format(k, v))
    update_str = ",".join(update_str)
    sql = 'update `{}` set {} where {};'.format(table_name, update_str, condition_str)
    exec_sql_script(sql_script=sql)


def read_dataframe_from_mysql(table_name=None, condition_dict: dict = None, result_col: list = None,
                              limit=10,
                              offset=10, cursor=None, **kwargs):
    """
    将mysql数据库表中的数据读入到pd.DataFrame类型结构中，默认全部读取，通过设置condition_dict添加过滤

    :param cursor: 连接了db文件的游标
    :param table_name: 数据库中表的名称
    :param condition_dict: 读取的数据条目所需满足的条件，默认全部读取
    :param result_col: 返回的数据列，默认全部返回
    :param limit: 返回条目的限制数量，同mysql数据库select语句中limit的用法
    :param offset: 返回条目的数据便宜，同mysql数据库select语句中offset的用法
    :param kwargs:
    :return:
    """
    cursor = _update_cursor(cursor)  # 更新模块变量last_cursor
    if isinstance(cursor, Cursor):
        con = cursor.connection
    elif isinstance(cursor, Engine):
        con = cursor
    data = pd.read_sql(f"select * from {table_name};", con=con)
    return data


def select_in_table(table_name=None, condition_dict: dict = None, result_col: list = None, limit=10,
                    offset=0,
                    fuzzy=False, result_type=None, cursor=None, **kwargs):
    """
    查

    精确查询，设置fuzzy为True，且condition_dict中的value为字符串值
    模糊查询，设置fuzzy为False，日期列不能使用模糊查询，只能使用范围查询
    范围查询，设置fuzzy为True，且condition_dict中的value为长度为2的列表，列表第一、二项分别为范围下、上限，且包含上下限

    当result_type=="json"时，返回的是一个list(dict)的json对象，即[{col1: value1, col2: value2,...}, ...}的json对象
    列表的每一项对应一条匹配的查询结果
    每一项的字典分别是{列名：值}

    kwargs={"date_format": "%Y-%m-%d %H:%M:%S"} 如果mysql中存在日期列，需要将日期转换为字符串，该参数定义日期字符串格式


    :param cursor:
    :param table_name:
    :param condition_dict:
    :param result_col: 不传入或传入空列表，则返回数据库中所有列
    :param limit:
    :param offset:
    :param fuzzy: 是否模糊查询
    :param result_type: 返回类型，如果为json，则返回为json格式的字符串
    :return: None或tuple，出错时返回None，如列明不存在等；否则返回一个tuple类型的数据，长度为0表示未查询到满足条件的数据
    """
    cursor = _update_cursor(cursor)  # 更新模块变量last_cursor
    if result_col is None:
        result_col = []
    condition_str = []
    result_col_str = []
    if fuzzy:
        for k, v in condition_dict.items():
            type_col = get_type_of_column(table_name=table_name, column_name=k)
            if type_col == "datetime":
                logger.error("日期类型不能使用模糊查询，请使用范围查询！")
                raise ValueError("日期类型不能使用模糊查询，请使用范围查询！")
            elif type_col == "int":
                logger.error("int类型不能使用模糊查询，请使用范围查询！")
                raise ValueError("int类型不能使用模糊查询，请使用范围查询！")
            elif type_col == "float":
                logger.error("float类型不能使用模糊查询，请使用范围查询！")
                raise ValueError("float类型不能使用模糊查询，请使用范围查询！")
            elif type_col is None:
                logger.warning(f"列类型未知，检查表中是否存在指定列({k})")
                return None
            condition_str.append(f"`{k}` LIKE '%{v}%'")
        condition_str = " and ".join(condition_str)
        condition_str = f"({condition_str})"
    else:
        for k, v in condition_dict.items():
            # 需要判断k是不是日期类型，日期类型需要将k,v都转成字符串进行对比
            type_col = get_type_of_column(cursor=cursor, table_name=table_name, column_name=k)
            if type_col is None:
                logger.warning(f"列类型未知，检查表中是否存在指定列({k})")
                return None
            elif type_col == "datetime":
                date_format = kwargs.get("date_format") or '%Y-%m-%d %H:%M:%S'
                if isinstance(v, datetime.datetime):  # 单值查询
                    v = datetime.datetime.strftime(v, date_format)
                    condition_str.append(f"DATE_FORMAT(`{k}`, '{date_format}')='{v}'")
                elif isinstance(v, datetime.date):
                    v = datetime.date.strftime(v, date_format)
                    condition_str.append(f"DATE_FORMAT(`{k}`, '{date_format}')='{v}'")
                elif isinstance(v, datetime.time):
                    v = datetime.time.strftime(v, date_format)
                    condition_str.append(f"DATE_FORMAT(`{k}`, '{date_format}')='{v}'")
                elif isinstance(v, str):
                    condition_str.append(f"DATE_FORMAT(`{k}`, '{date_format}')='{v}'")
                elif isinstance(v, list):  # 如果是列表，则说明是日期范围
                    if isinstance(v[0], datetime.datetime):  # 如果是datetime对象，则需要转换为str
                        v[0] = datetime.datetime.strftime(v[0], date_format)
                        v[1] = datetime.datetime.strftime(v[1], date_format)
                    elif isinstance(v[0], datetime.date):
                        v[0] = datetime.date.strftime(v[0], date_format)
                        v[1] = datetime.date.strftime(v[1], date_format)
                    elif isinstance(v[0], datetime.time):
                        v[0] = datetime.time.strftime(v[0], date_format)
                        v[1] = datetime.time.strftime(v[1], date_format)
                    condition_str.append(f"DATE_FORMAT(`{k}`, '{date_format}') "
                                         f"between '{v[0]}' and '{v[1]}'")
            elif type_col == "int" or type_col == "float":  # 如果对应列是数值类型
                if isinstance(v, list):  # 范围查询
                    condition_str.append(f"`{k}` between {v[0]} and {v[1]}")
                else:  # 单值查询
                    condition_str.append(f"`{k}`={v}")
            else:
                if isinstance(v, list):  # 范围查询
                    condition_str.append(f"`{k}` between '{v[0]}' and '{v[1]}'")
                else:  # 单值查询
                    condition_str.append("`{}`='{}'".format(k, v))
        condition_str = " and ".join(condition_str)
    if len(result_col) == 0:
        result_col_str = "*"
    else:
        for col in result_col:
            result_col_str.append(f"`{col}`")
        result_col_str = ",".join(result_col_str)

    sql = 'select {} from `{}` where {} limit {} offset {};'.format(result_col_str, table_name, condition_str, limit,
                                                                    offset)
    # 返回默认格式的检索结果
    result = exec_sql_script(sql_script=sql)
    # 返回json格式的检索数据，是一个列表，每一个列表项对应一条匹配的记录
    if result_type == "json":  # 如果需要返回json格式结果
        date_format = kwargs.get("date_format") or '%Y-%m-%d %H:%M:%S'
        if len(result) > 0:
            temp_dicts = []
            if result_col_str == "*":
                result_col = get_column_name_of_table(cursor=cursor, table_name=table_name)
            for item in result:
                temp_dict = {}
                for i, col in enumerate(result_col):
                    # 如果列是日期类型，需要转为字符串返回，否则json.dumps()方法会报错
                    if get_type_of_column(table_name, col) == "datetime":
                        if item[i] is not None:  # 如果是非必须的时间列，则时间值可能为None
                            temp_dict.update({col: item[i].strftime(date_format)})
                        else:
                            temp_dict.update({col: None})
                    else:
                        temp_dict.update({col: item[i]})
                temp_dicts.append(temp_dict)
            result = json.dumps(temp_dicts, ensure_ascii=False)
        else:
            result = json.dumps({})
    return result


def get_reference_table_column(table_name=None, column_name=None, cursor: Cursor = None):
    """
    根据已知表名和列名查找其外键指向的父表和父表列名

    :param cursor:
    :param table_name: 当前表名
    :param column_name: 当前列名
    :return: 外键指向的所有父表和父表列名的tuple
    """
    # sql = "select REFERENCED_TABLE_NAME,COLUMN_NAME from INFORMATION_SCHEMA.KEY_COLUMN_USAGE where" \
    #       " TABLE_NAME = '{}'and COLUMN_NAME='{}' and CONSTRAINT_NAME!='PRIMARY';".format(tableName, columnName)
    # results = exec_sql_script(cursor=cursor, sql_script=sql)
    pass
    return None


def add_foreign_key(table_master=None, key_master=None, table_follow=None, key_follow=None, cursor=None):
    """
    添加外键约束

    适用于一对多关系，一是主表，多是从表，

    :param cursor:
    :param table_follow:
    :param key_follow:
    :param table_master:
    :param key_master:
    :return:
    """
    cursor = _update_cursor(cursor)  # 更新模块变量last_cursor
    constraint_name = f"c_{table_master}_{table_follow}"
    if has_table(table_follow, cursor):  # constraint {constraint_name}
        sql = f"alter table `{table_follow}` add " \
              f"foreign key (`{key_follow}`) references `{table_master}`(`{key_master}`);"
        exec_sql_script(cursor=cursor, sql_script=sql)


def create_mysql_relation_table(table1=None, column1=None, table2=None, column2=None, cursor=None):
    """
    创建关系表

    :param cursor: 游标
    :param table1: 表1
    :param column1: 表1的主键
    :param table2: 表2
    :param column2: 表2的主键
    :return:
    """
    cursor = _update_cursor(cursor)  # 更新模块变量last_cursor
    table_r = f"{table1}_rl_{table2}"
    column1_name_in_this_table = f"{table1}_{column1}"
    column2_name_in_this_table = f"{table2}_{column2}"
    type1 = get_type_of_column(table1, column1)
    type2 = get_type_of_column(table2, column2)

    if not has_table(table_r, cursor):
        sql = f"create table `{table_r}`(" \
              f"id int primary key auto_increment, " \
              f"`{column1_name_in_this_table}` {type1} not null, " \
              f"`{column2_name_in_this_table}` {type2} not null, " \
              f"foreign key(`{column1_name_in_this_table}`) references `{table1}`(`{column1}`) ON DELETE " \
              f"CASCADE ON UPDATE CASCADE, " \
              f"foreign key(`{column2_name_in_this_table}`) references `{table2}`(`{column2}`) ON DELETE " \
              f"CASCADE ON UPDATE CASCADE, " \
              f"unique key(`{column1_name_in_this_table}`,`{column2_name_in_this_table}`));"
        exec_sql_script(cursor=cursor, sql_script=sql)


def get_type_of_column(table_name=None, column_name=None, cursor=None):
    """
    获取mysql表中字段的类型，如果不设置column_name则返回所有的字段类型
    :param cursor:
    :param table_name:
    :param column_name: 为空则依次返回所有列的类型，封装为一个列表
    :return:
    """
    cursor = _update_cursor(cursor)  # 更新模块变量last_cursor
    try:
        results = exec_sql_script(sql_script=f"show columns from `{table_name}`;")
    except pymysql.err.ProgrammingError:
        logger.error(f"表名或字段名不存在！'{table_name}.{column_name}'")
        return None
    type_dict = {}
    if column_name is None:
        for result in results:
            type_dict[result[0]] = result[1].lower()
        return type_dict
    else:
        for result in results:
            if result[0] == column_name:
                return result[1].lower()
    logger.error(f"表名或字段名不存在！'{table_name}.{column_name}'")
    return None


def get_column_name_of_table(table_name=None, cursor=None):
    """
    获取表中所有的字段名

    :param cursor:
    :param table_name:
    :return: list
    """
    cursor = _update_cursor(cursor)  # 更新模块变量last_cursor
    sql = f"select COLUMN_NAME from information_schema.COLUMNS where table_name='{table_name}';"
    res = cursor.execute(sql)
    if isinstance(cursor, Cursor):
        fetchall = cursor.fetchall()
    else:
        fetchall = res.fetchall()
    result = []
    for field in fetchall:
        result.append(field[0])
    return result


def generate_columns_values(info_dict: dict, column_alias_dict: dict,
                            parameters: list = None,
                            extras_field_name: str = "extras_field",
                            null_symbol: list = ['-', '', '/', '－', '—']):
    """
    根据用户自定义的信息字典和数据库中字段的别名字典，生成可以填充到mysql数据库表格中的columns和values，以便使用
    INSERT INTO table_name (列1, 列2,...) VALUES (值1, 值2,....)
    语句进行数据库数据条目插入。

    :param null_symbol: 代表值为空的字符串列表
    :param extras_field_name: 当用户传入的变量找不到对应的字段名时，会将该列及其值以字典形式更新进extras_field_name对应的数据库表字段值中
    :param parameters: 需要过滤单位的数据库表字段的列表，有些用户传入变量带有单位，如果不过滤会导致类型不匹配，如0.6MPa不能转为float类型
    :param info_dict: 表格中解析得到的变量名及其对应的数值，如{"生日": "19910305"}，这里的变量名不同来源可能不同，但指的是同一个东西
    :param column_alias_dict: 字段的别名，如{"birthday": ["生日", "出生日期"]}，这里指出了"birthday"字段的所有可能的别名
    :return: [列1, 列2,...], [值1, 值2,...]，例如["birthday"], ["19910305"]，如果为值为空，则返回结果中不包括该字段列
    """

    def find_col_name_in_db(name_variety: str):
        """
        用于获取用户变量在数据库中对应的字段名。

        column_alias_dict是字段对应的别名的字典。其格式如下：
        {
        "p_d_rhs1": ["额定再热汽门前压力", "再热蒸汽压力"],  # p_d_rhs1是数据库中的字段名，右侧列表是用户可能使用的变量名
        "t_d_rhs1": ["额定再热汽门前温度", "再热蒸汽温度"],  #
        }

        因为用户在实际word文档中，同一个变量可能使用不同的名字，这里将其对应到同一个数据库字段上。如果没有找到用户在word中定义的变量的存储字段，
        则该方法返回"extras_field"

        :param name_variety: 用户在word中使用的变量名
        :return: 数据库库中存储的变量名
        """
        for name_table, alias_list in column_alias_dict.items():
            if name_variety in alias_list:
                return name_table
        return extras_field_name

    columns_names = []
    columns_values = []
    extras_field_values = {}
    for k, v in info_dict.items():
        name = find_col_name_in_db(k)
        if name != extras_field_name:
            if name in parameters:
                if v.strip().endswith("%"):
                    v: str = v.replace("%", "").strip()
                    try:
                        v: float = float(v) / 100
                    except ValueError:
                        logger.warning(f"数值错误，错误字段为：{k}={v}，错误表格为：")
                        logger.warning(json.dumps(info_dict, ensure_ascii=False))
                        continue  # 出错了则略过该字段
                elif v.strip() in null_symbol:
                    continue  # 变量值为空则略过该字段
                else:
                    v = re.findall("([0-9\.]*)", v)[0]
            columns_names.append(name)
            columns_values.append(v)
        else:
            extras_field_values.update({k: v})
    if len(extras_field_values) > 0:
        columns_names.append("extras_field")
        columns_values.append(json.dumps(extras_field_values, ensure_ascii=False))
    return columns_names, columns_values


def get_update_time_of_table(table_name, or_create_time=False, cursor=None):
    """
    获取表的最后更新时间

    :param cursor: Cursor or Engine
    :param table_name:
    :param db: 表所属的db，不设置默认在cursor绑定的db中查找
    :param or_create_time: 当表的更新时间为None时，如果该参数为False,则返回None，否则返回表的创建时间，需要注意的是创建时间也可能是None
    :return: datetime.datetime类型的时间或None
    """
    # mysql的更新时间使用该方法是无法获得的，需要后期mysql版本更新看能否解决
    # sql = f"SELECT update_time FROM information_schema.tables WHERE table_name='{table_name}'"
    # fetch = exec_sql_script(cursor=cursor, sql_script=sql )
    # if len(fetch) == 0:
    #     return None
    # else:
    #     if or_create_time:
    #         return get_create_time_of_table(cursor, table_name)
    #     else:
    #         return fetch[0][0]
    # 暂时使用专门的modifyTable表记录每个表的更新时间
    _update_cursor(cursor)  # 更新模块变量last_cursor
    if not has_table("modifytime"):  # Sqlite数据库也支持
        return None
    else:
        result = select_in_table("modifytime", condition_dict={"table": table_name}, result_col=['datetime'])
        if len(result) == 1:
            return result[0][0]
        elif len(result) == 0:
            return None
        else:
            logger.error(f"查找到多个表名匹配{table_name}的表，无法确定需要哪个表的更新时间")


def update_update_time_of_table(table_name: str = None, cursor: Cursor or Engine = None):
    """
    更新表的更新时间

    :param cursor:
    :param table_name:
    :return:
    """
    _update_cursor(cursor)  # 更新模块变量last_cursor
    if not has_table("modifyTime"):
        meta_data = MetaData()  # 没有关系的表放在不同的metadata中
        table = Table('modifyTime', meta_data,
                      Column('table', String, primary_key=True),
                      Column('datetime', DateTime, nullable=False))
        # create_table("modifyTime", columns={"table": "varchar(50)", "datetime": "datetime"})
        create_table("modifyTime", metadata=table)
    update_time = datetime.datetime.now()
    insert_item_to_mysql("modifyTime", [table_name, update_time], replace=True)


def get_create_time_of_table(table_name=None, cursor: Cursor = None):
    """
    查询表的创建日期

    :param cursor:
    :param table_name:
    :return:
    """
    cursor = _update_cursor(cursor)  # 更新模块变量last_cursor
    sql = f"SELECT create_time FROM information_schema.tables WHERE table_name='{table_name}'"
    fetch = exec_sql_script(cursor=cursor, sql_script=sql)
    if len(fetch) == 0:
        return None
    else:
        return fetch[0][0]


# 测试的mysql和adminer版本分别为：8.0.19、4.7.6
if __name__ == '__main__':
    if not start_mysql_service():
        os.system("pause")

# 通过命令 pyinstaller -n mysql -i yangke2.ico -F mysql.py 生成exe文件



import os.path
import re
import traceback
from typing import Iterable

from yangke.common.QtImporter import QModelIndex, QMimeData, QComboBox, QApplication, QLineEdit, QDrag

from yangke.common.QtImporter import QDir, Qt, QRectF, QSize, QStringListModel, QStandardItemModel, QStandardItem, \
    QIcon, QColor, QBrush, QResizeEvent, QKeyEvent, QFileSystemModel, QListView, QWidget, QGridLayout, QVBoxLayout, \
    QListWidget, QListWidgetItem, QLabel, QLayout, uic, QMainWindow, QColorDialog

from UI.MainWindow_PyQt5 import Ui_MainWindow
from UI.content import Ui_Form as ContentForm
from yangke.base import yield_all_file
from yangke.common.config import logger
from yangke.common.qt import UIWidget, run_app, YkWindow, YkItem, YkFoldableWidget, YkInputPanel, YkEditor, YkConsole, \
    YkVariableWidget
from yangke.ebsilon.constant.constant import default_unit, Unit
from yangke.ebsilon.graphicsview import YkGraphicsScene, YkGraphicsView, YkGraphicsItem, CoordinateItem, SceneGridItem, \
    YkStyledItemDelegate, DisplayPanel, PipeItem, SceneFlagItem, Port
import importlib

from yangke.ebsilon.values import Values


class Content(QWidget):
    def __init__(self, tab_name):
        super(Content, self).__init__()
        self.ui: QWidget = UIWidget(ContentForm)  # 即qt designer设计的ui文件面板
        self.tab_name = tab_name
        self.setLayout(QGridLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        # noinspection all
        self.layout().addWidget(self.ui, 0, 0, 1, 1)
        self.scene = YkGraphicsScene()
        self.view = YkGraphicsView(self.scene)  # self.view.setScene(self.scene)
        self.view.setAcceptDrops(True)
        # self.view.setDragMode(ScrollHandDrag)
        self.view.setAlignment(Qt.AlignCenter)
        self.ui.layout().addWidget(self.view)


class CalResult:
    def __init__(self):
        ...


class MainWindow(YkWindow, QMainWindow):
    def __init__(self):
        super(MainWindow, self).__init__()
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.choose_component = 0
        self.setWindowTitle("热力性能建模平台")
        self.ui.btn_set_value.clicked.connect(self.set_values_panel2data)
        self.ui.btn_cancel.clicked.connect(self.cancel)
        # # self.ui下有 self.ui.info_tab用于显示组件的详细信息
        self.ui.sfv.setLayout(QVBoxLayout())  # sfv的竖直方向的SizePolicy必须是Fixed
        self.ui.res.setLayout(QVBoxLayout())
        self.ui.mat.setLayout(QVBoxLayout())

        # [<PyQt5.QtGui.QStandardItem object at 0x0227C49C11B0>, <PyQt5.QtGui.QStandardItem object at 0x0227C49C1480>]
        self.items_nav = []  # 组件的QStandardItem实例，是导航面板上的显示对象，长度与self.all_comps相同

        # 组件类对象, {'Comp1': <class 'components.Comp1.Item'>, 'Comp2': <class 'components.Comp2.Item'>}
        self.all_comps = self.load_all_comp(os.path.join(os.path.dirname(__file__), "components"))
        self.init_file_view()
        self.init_compo_view()
        self.content_widget = self.init_content_tab()

        # 设置场景导航面板的数据模型位self.lv_scene_model
        self.lv_scene_model = QStandardItemModel()
        self.update_scene_nav_view(None)

        self.ui.splitter_3.setStretchFactor(0, 2)
        self.ui.splitter_3.setStretchFactor(1, 4)
        self.current_comp_info_tab = "SpecificationValues"  # 表示self.ui.info_tab显示按个数据面板
        self.ui.content_tab.tabCloseRequested.connect(self.close_project)  # 使单个项目可以关闭

        if self.proj.get("DefaultUnit") is not None:  # 项目默认的单位系统不为空，则加载项目默认单位系统
            pass
        elif self.ini_info.get("DefaultUnit") is not None:  # 如果项目单位系统为空，但软件单位系统不为空，则加载软件单位系统
            self.proj["DefaultUnit"] = self.ini_info.get("DefaultUnit")
        else:  # 如果以上二者都为空，则初始化一个默认的单位系统，并保存到项目中
            self.proj["DefaultUnit"] = None
        if self.proj["DefaultUnit"] is not None:
            default_unit.set(self.proj["DefaultUnit"])

        if self.proj_file is not None and os.path.exists(self.proj_file):  # 如果有项目文件，且该文件存在
            try:
                self.open(self.proj_file)
                self.close_project(0)
            except:
                logger.warning(f"打开{self.proj_file}时出错")

        self.exec_scope = {"self": self}  # 终端代码执行时的变量作用域
        self.tv_var = YkVariableWidget()  # 固定面板-> Terminal -> 变量空间
        self.ui.verticalLayout_12.addWidget(self.tv_var)
        self.init_fixed_window()

        self.ui.info_tab.currentChanged.connect(self.info_tab_changed)
        self.status_bar_coor = QLabel('')
        self.status_bar_coor.setFixedWidth(60)
        self.statusBar1.addPermanentWidget(self.status_bar_coor)
        self.calc_result = CalResult()  # 计算结果信息，如计算是否成功，计算失败原因等

    def keyPressEvent(self, a0: QKeyEvent) -> None:
        if a0.key() == Qt.Key_F9:
            self.calculate()
        elif a0.modifiers().value == (Qt.ShiftModifier | Qt.ControlModifier).value and a0.key() == Qt.Key_AsciiTilde:
            if self.ui.dockWidget_ATW.isHidden():
                self.ui.dockWidget_ATW.show()
            else:
                self.ui.dockWidget_ATW.hide()
        else:
            super(MainWindow, self).keyPressEvent(a0)

    def close_project(self, sender=None):
        if isinstance(sender, int):
            self.ui.content_tab.removeTab(sender)
            self.content_widget = self.ui.content_tab.currentWidget()
        else:
            logger.warning("非法索引！")

    def init_content_tab(self):
        if self.ui.content_tab.count() == 0:
            return self.add_content_tab1("新建项目1")
        else:
            self.content_widget = self.ui.content_tab.widget(0).ui
            self.content_widget.scene.item_changed_signal.connect(self.update_scene_nav_view)
            return self.content_widget

    def add_content_tab1(self, tab_name):
        """
        不能使用父类的add_content_tab方法。
        """
        self._content_tab = self.ui.content_tab
        if tab_name in self.get_content_tab_labels():
            self._content_tab.setCurrentIndex(self.get_content_tab_labels().index(tab_name))
            logger.debug(f"已经存在名为{tab_name}的项目")
            self.content_widget = self._content_tab.currentWidget()
        else:
            content = Content(tab_name)
            self.ui.content_tab.addTab(content, tab_name)
            self.content_widget = content
            self.content_widget.scene.setBackgroundBrush(QBrush(QColor(208, 208, 208)))
            self.content_widget.scene.item_changed_signal.connect(self.update_scene_nav_view)
        return self.content_widget

    def load_all_comp(self, path):
        """
        从yangke.ebsilon.components文件夹中加载所有的组件

        :param path:
        :return:
        """
        if hasattr(self, "all_comps") and self.all_comps is not None:
            return self.all_comps
        all_comps = {}
        for file in yield_all_file(path, ".py"):
            basename = os.path.basename(file).replace(".py", "")
            _ = importlib.import_module(f"components.{basename}")
            all_comps.update({basename: _.Item})
        return all_comps

    def init_compo_view(self):
        """
        组件导航面板，加载及显示所有组件

        :return:
        """
        list_model = QStandardItemModel()
        self.items_nav = []
        for k, comp_cls in self.all_comps.items():
            comp: YkGraphicsItem = comp_cls()
            if comp.icon is None:
                comp.generate_icon()

            item = QStandardItem(comp.icon, comp.NAME)
            item.setData({"ebs_id": comp.EBS_ID, "ebs_name": comp.EBS_NAME, "ebs_type": comp.EBS_TYPE},
                         Qt.UserRole)  # 给item添加自定义数据
            self.items_nav.append(item)
            item.setDragEnabled(True)  # 是列表中的项可以被拖动
            list_model.appendRow(item)

        # list_model.mimeData()  # 实现list_model的mimeData()使之可以
        def _mimeData(indexes: list[QModelIndex]):
            if len(indexes) <= 0:
                return 0
            data = QMimeData()
            data.setData("com_data", str(indexes[0].data(Qt.UserRole)).encode("utf8"))
            return data

        list_model.mimeData = _mimeData

        self.ui.lv_com.setItemDelegate(YkStyledItemDelegate())
        self.ui.lv_com.setSpacing(4)

        self.ui.lv_com.setModel(list_model)
        self.ui.lv_com.setViewMode(QListView.ListMode)
        self.ui.lv_com.setDragEnabled(True)

        self.ui.lv_com.clicked.connect(self.click_com_cls)

    def init_file_view(self):
        """
        文件目录面板
        :return:
        """
        model = QFileSystemModel()
        model.setRootPath(QDir.currentPath())
        self.ui.tv_file.setModel(model)

    def update_scene_nav_view(self, idx: str | None = None):
        """
        场景导航面板

        :param idx: 新增设备的场景id，每次新增设备组件时，会触发本方法执行
        :return:
        """
        if idx is None:  # 初始化场景面板组件时，传入的id位scene
            self.ui.lv_scene.setSpacing(4)
            self.ui.lv_scene.setModel(self.lv_scene_model)
            self.ui.lv_scene.setViewMode(QListView.IconMode)
            self.ui.lv_scene.setDragEnabled(True)
            self.ui.lv_scene.clicked.connect(self.click_com_obj)

        elif idx.startswith("del"):
            name = idx.replace("del", "").strip()
            Qt.FindChildrenRecursively = 1
            # item = self.lv_scene_model.findChild(name=name, options=Qt.FindChildrenRecursively)
            row = self.lv_scene_model.findItems(name)[0].row()
            self.lv_scene_model.removeRow(row)
            self.ui.lv_scene.clearFocus()
            self.ui.lv_scene.clearSelection()
        else:
            scene_id = int(idx)
            设备组件: YkGraphicsItem = self.content_widget.scene.items_scene_id.get(scene_id)
            icon = 设备组件.icon if hasattr(设备组件, "icon") else None
            icon = icon or 设备组件.generate_icon()
            name = 设备组件.NAME if hasattr(设备组件, "NAME") else None
            ebs_id = 设备组件.EBS_ID if hasattr(设备组件, "EBS_ID") else None
            ebs_name = 设备组件.EBS_NAME if hasattr(设备组件, "EBS_NAME") else None
            ebs_type = 设备组件.EBS_TYPE if hasattr(设备组件, "EBS_TYPE") else None

            item = QStandardItem(icon, f"{name}_{设备组件.id} / S{scene_id}")
            item.setData({
                "ebs_id": ebs_id, "ebs_name": ebs_name, "ebs_type": ebs_type, "scene_id": scene_id
            }, role=Qt.UserRole)  # 给item添加自定义数据
            self.lv_scene_model.appendRow(item)
        # self.ui.lv_com.setItemDelegate(YkStyledItemDelegate())

    def init_fixed_window(self):
        self.ui.Terminal.setLayout(QVBoxLayout())
        self.ui.Terminal.layout().setContentsMargins(0, 0, 0, 0)
        self.ui.terminal = YkConsole()
        self.ui.Terminal.layout().addWidget(self.ui.terminal)
        self.ui.Log.setLayout(QVBoxLayout())
        self.ui.Log.layout().setContentsMargins(0, 0, 0, 0)
        self.ui.log_editor = YkEditor()
        self.ui.Log.layout().addWidget(self.ui.log_editor)
        self.ui.terminal.deal_cmd = self.deal_cmd
        self.ui.dockWidget_ATW.hide()
        self.update_terminal_var()
        logger.add(self.ui.log_editor)  # 将日志面板加入日志类的sink中

    def update_terminal_var(self):
        """
        更新变量面板上的变量值
        """
        tv = self.tv_var  # 变量的QTreeView实例对象
        try:
            self.exec_scope.update({"items_scene_id": self.content_widget.scene.items_scene_id})
            tv.update_variable(self.exec_scope)
        except:
            self.output_log(traceback.format_exc())

    def deal_cmd(self, cmd):
        """
        处理控制台输入的命令

        :param cmd: 控制台最后一行的文本内容，也就是命令
        :return:
        """
        try:
            if cmd == "refresh" or cmd == "update":  # 更新一下当前数据
                self.update_terminal_var()
                return
            res = eval(cmd, self.exec_scope)  # eval只能执行表达式，不能是语句，如eval("x=1")会报错
            self.exec_scope.update({"_": res})
        except:
            try:
                res = exec(cmd, self.exec_scope)  # exec可以执行语句，但是执行表达式时不返回结果，如exec("2+1")返回None
            except:
                res = traceback.format_exc()
        self.update_terminal_var()
        return res

    def click_com_cls(self, item):
        """
        组件导航面板的点击响应事件

        :return:
        """
        user_info = item.data(Qt.UserRole)
        self.choose_component = user_info.get("ebs_id")

    def insert_components(self, com_name):
        """
        在当前激活的项目中插入组件，一般为text组件
        """
        if com_name == 'text':
            self.choose_component = 'text'

    def click_com_obj(self, item):
        """
        场景导航面板的点击响应事件

        :param item:
        :return:
        """
        user_info = item.data(Qt.UserRole)
        if user_info is None:
            return
        scene_id = user_info.get("scene_id")
        item: YkGraphicsItem = self.content_widget.scene.items_scene_id.get(scene_id)
        item.setFocus(Qt.MouseFocusReason)
        self.display_info(scene_id)

    def draw_content_view(self):
        if self.content_widget is None:
            return
        self.content_widget.scene.setBackgroundBrush(QBrush(QColor(0, 100, 0, 50)))

    def resizeEvent(self, a0: QResizeEvent) -> None:
        super(MainWindow, self).resizeEvent(a0)
        # try:
        #     rect = self.content_widget.rect()
        #     self.content_widget.scene.setSceneRect(QRectF(rect.x(), rect.y(), rect.width() - 24, rect.height() - 24))
        # except:
        #     pass

    def save(self):
        """
        保存项目
        :return:
        """
        self.proj = self.proj or {}
        items = self.content_widget.scene.items_scene_id
        # 将items中无法序列化的Item对象转换为字符串
        items_new = []
        try:
            for scene_id, obj in items.items():
                if scene_id <= 2 or isinstance(obj, CoordinateItem) or isinstance(obj, SceneGridItem):
                    continue  # scene_id=0和1是网格和坐标系组件
                obj: YkGraphicsItem = obj
                bak_info = {}
                if obj.EBS_ID == '-100':
                    obj: PipeItem = obj
                    bak_info.update({
                        "control_points": obj.control_points, "color": obj.color, "connect_items": obj.connect_items,
                    })
                objs_pickable = {"EBS_ID": obj.EBS_ID, "EBS_NAME": obj.EBS_NAME, "id": obj.id, "scene_id": obj.scene_id,
                                 "values": obj.values, "rotation": obj.rotation(),
                                 "scene_pos": obj.scenePos(), "height": obj.height, "width": obj.width,
                                 "ports": obj.ports, "bak_info": bak_info, "z_value": obj.zValue(),
                                 }
                items_new.append(objs_pickable)
        except:
            traceback.print_exc()

        self.proj.update({'items': items_new})
        super(MainWindow, self).save()

        self.statusBar1.showMessage(f"项目保存完成")
        logger.info(f"项目保存至{self.proj_file}")

    def get_projects(self):
        """
        获取当前已经打开的项目名称
        """
        res = [self.ui.content_tab.tabText(i) for i in range(self.ui.content_tab.count())]
        return res

    def new_project(self):
        super().new_project()
        _proj_name = "new"
        _proj_names = self.get_projects()
        if _proj_name in _proj_names:
            i = 1
            while f"new_{i}" in _proj_names:
                i = i + 1
            _proj_name = f"new_{i}"
        _ = self.add_content_tab1(_proj_name)
        self.ui.content_tab.setCurrentWidget(_)
        self._content_tab = self.ui.content_tab

    def open(self, proj=None):
        super(MainWindow, self).open(proj)
        # 添加一个新的content_tab，并切换到新添加的content_tab
        if self.proj_file is None:  # 当新建项目时，打开空项目但proj_file尚未指定
            return
        proj_name = os.path.basename(self.proj_file).split(".")[0]
        self.add_content_tab1(tab_name=proj_name)
        self.ui.content_tab.setCurrentWidget(self.content_widget)
        if self.proj is None:
            return
        # 取出项目中的组件，并逐个绘制到场景中
        proj_items = self.proj.get("items")
        if proj_items is None:
            return
        try:
            for obj in proj_items:
                EBS_ID = obj.get("EBS_ID")
                if EBS_ID == '-100':
                    comp: PipeItem = PipeItem(fluid_type=obj["EBS_NAME"])
                    comp.set_color(obj["bak_info"]["color"])
                    comp.setPos(obj.get("scene_pos"))
                    control_points = obj["bak_info"]["control_points"]
                    comp.update_line(vectors_item=control_points)
                    comp.set_connect_items(obj["bak_info"]["connect_items"])
                    comp.EBS_NAME = obj["EBS_NAME"]
                else:
                    comp: YkGraphicsItem = self.all_comps.get(f"Comp{obj.get('EBS_ID')}")()
                    comp.setPos(obj.get("scene_pos"))
                    comp.values = obj.get("values")
                    comp.ports = obj.get("ports")

                z_value = obj.get("z_value") or 1  # 图层高度
                comp.setZValue(z_value)
                comp.setRotation(obj.get("rotation"))  # 图像的旋转角度
                self.content_widget.scene.addItem(comp, scene_id=obj.get("scene_id"), id=obj.get('id'))  # 添加combobox项时
            logger.debug(f"项目加载完成")
        except:
            traceback.print_exc()

        self.update_scene_nav_view()

    def refresh(self):
        ...

    def get_calc_items(self, sort_by='计算顺序'):
        """
        获取当前激活的项目下需要参与计算的设备组件字典
        """
        components = []
        for scene_id, com in self.content_widget.scene.items_scene_id.items():
            if com.__class__.__name__ in ["SceneGridItem", "SceneFlagItem", "CoordinateItem"]:
                ...
            else:
                components.append(com)
        if sort_by == "计算顺序":
            components.sort(key=lambda x: x.cal_order)
        return components

    def clear_results(self, components=None):
        """
        清空当前激活的项目中的计算结果数据
        """
        components = components or self.get_calc_items()
        for com in components:
            com: YkGraphicsItem = com
            for idx, port in com.ports.items():
                port: Port = port
                port.values = Values(values_steps={0: {}})  # port的初始值为空
            com.values.clear_result()  # 清空端口上初始值以后的迭代步的值
            com.calculated = False  # 重置计算状态为False

    def calculate_real_time(self):
        ...

    def calculate(self):
        """
        计算项目，计算开始时，应确保每个组件的values中的values_steps数组中已经放置了初始的参数设置，即使没有初始参数需要设置，也需要以
        空Values对象占位。

        计算过程中，需要保证时间步0只保存画面上传入的数据，中间计算结果永远不要保存到时间步0。
        计算开始时，所有的Port对象中均没有数据，Port上的数据是在事件步1由边界条件设置组件传递到Port对象上的

        :return:
        """
        cal_type, _info = self.get_calculate_info()
        if cal_type == "real_time":  # 实时计算，每隔指定时间计算一次
            time_interval = float(_info[0])
            from yangke.base import execute_function_by_interval
            execute_function_by_interval(self.calculate_real_time, minute=0, second=time_interval)
        else:
            components = self.get_calc_items()  # 所有参与计算的设备组件字典
            epochs = 100  # 计算迭代总次数
            exit_msg = "计算失败！"
            error_info = {}

            # 正式计算之前，清空所有组件的port中存储的之前的计算结果以及result中保存的计算结果
            self.clear_results(components)

            # 首先计算边界值组件，并将边界值组件的值传递给相应的端口，边界值组件必须可以独立计算，否则会直接退出计算过程
            # 边界值组件上设置的参数也是初始参数，需要传递给port的value_steps第0步
            for com in components:
                if hasattr(com, "EBS_ID") and com.EBS_ID in [1, 46]:
                    msg: str = com.calculate(step=0)
                    if msg != "done":
                        logger.debug(f"{com.scene_id}号组件未设置必要参数！--- {msg}")
                        epochs = 1  # 让整体计算不在进行，直接退出
                        break
                    com.spread_values(step=0)  # 将当前组件设置的边界或起始值扩散到与其端口上

                    com.calculate(step=1)
                    com.spread_values(step=1)  # 将当前组件设置的边界或起始值扩散到与其端口上

            for i in range(1, epochs):
                total_num = len(components)
                calculated = {}
                exit_epochs = False
                self.content_widget.scene.current_step = i  # 设置当前计算步
                for com in components:  # 不加dict，调试经常出现无法解析整形的错误
                    if com.EBS_ID in [1, 46]:  # 边界及初始条件设置组件已经计算完成
                        # 目前来讲，1和46号组件无须重复计算，但是考虑到初始值或参考值或控制值等组件，每个迭代步可能都要计算
                        # 且如果这里不计算，则边界值组件的values.value_steps只有前两步的值
                        com.calculate(step=i)
                        com.spread_values(step=i)  # 将当前组件设置的边界或起始值扩散到与其端口上
                        com.calculated = True
                        calculated.update({com.scene_id: True})
                        continue

                    # if com.calculated:  # 计算过的参数仍需计算，不影响过定义，因为过定义只在第1个时间步判断第0个时间步的值
                    #     calculated.update({scene_id: True})
                    #     continue

                    # 首先将当前面板上的数据读入设备组件的values中
                    com.before_com_calculate(step=i)
                    msg: str = com.calculate(step=i)
                    if msg is None:
                        logger.debug(f"组件{com.scene_id}的calculate方法未返回计算结果信息，请检查组件代码！")
                        return

                    msgs = msg.split("---")
                    _err = None
                    if len(msgs) == 2:
                        msg, _err = msgs[0].strip(), msgs[1].strip()
                    if msg == "done":
                        com.calculated = True
                        calculated.update({com.scene_id: True})
                    elif _err is not None and _err.strip() == "break":  # 引发计算立即退出的错误
                        error_info.update({com.scene_id: msgs})
                        logger.debug(error_info)  # 输出错误原因
                        exit_epochs = True
                        exit_msg = "计算失败！"
                        break
                    else:
                        com.calculated = False
                        if error_info.get(com.scene_id) == msg:  # 当前错误信息与上一次相同，则退出计算
                            logger.debug(error_info)
                            exit_epochs = True
                            exit_msg = "计算失败！"
                            break
                        error_info.update({com.scene_id: msg})  # 更新当前组件的错误信息
                        if len(calculated) >= total_num - 1:  # 其他都已经计算完成，只剩一个未计算，则说明缺少参数，直接退出并报错
                            logger.debug(msg)
                            exit_epochs = True
                            exit_msg = "计算失败！"
                            break
                    com.spread_values(step=i)  # 将当前组件计算出的结果扩散到与其连接的其他组件

                if len(calculated) == total_num:
                    # 如果所有组件都计算完成，则退出计算循环
                    exit_epochs = True
                    exit_msg = "计算成功"
                if exit_epochs:
                    break

            # 获取当前组件信息面板是属于哪个组件的，更新显示结果
            _ = self.ui.detail_tab.tabText(0).split("/ S")
            if len(_) > 1:
                scene_id_current = int(_[1].replace(']', '').strip()) if len(_) > 1 else None
                self.display_info(scene_id=scene_id_current, tab=self.current_comp_info_tab)

            logger.debug(exit_msg)
            self.ui.ActiveToolWindow.setCurrentWidget(self.ui.Log)  # 将FixedToolWindow切换到日志面板
            self.show_active_tool_window()

    def show_active_tool_window(self):
        if self.ui.dockWidget_ATW.isHidden():
            self.ui.dockWidget_ATW.show()

    def display_info(self, scene_id, tab=None, save_data=True):
        """
        当选中组件时，该方法用于显示组件的详细信息

        @param: scene_id : 组件在场景中的id
        @param tab: 需要显示的信息面板标签页
        """
        if save_data:
            self.set_values_panel2data()  # 先保存当前面板的值，后切换到新面板

        if scene_id is None or scene_id < 2:  # 0号组件是场景网格对象，1号组件是场景坐标系对象
            return True
        item = self.content_widget.scene.items_scene_id.get(scene_id)

        if item is None:
            return

        if tab is None:
            self.ui.detail_tab.setTabText(0, f"组件信息[{item.NAME}_{item.id} / S{item.scene_id}]")
            tab = self.ui.info_tab.tabText(self.ui.info_tab.currentIndex())

        self.show_info_tab(tab, item)

        # 更新场景组件导航面板上的组件列表
        # self.update_scene_nav_view()  # 貌似不用更新场景信息面板，否则会导致场景信息面板重加载

        idx = list(self.content_widget.scene.items_scene_id.keys()).index(scene_id) - 3  # 因为场景中有三个组件不显示
        qm_idx = self.ui.lv_scene.model().index(idx, 0)
        self.ui.lv_scene.setCurrentIndex(qm_idx)

    def show_info_tab(self, tab_name: str, device_item: YkGraphicsItem | None = None):
        """
        根据tab_items的值，构建组件信息面板，并将values中的值显示到信息面板上

        :param tab_name: 信息面板的标签页名
        :param device_item: 当面板上的combobox需要链接事件时，需要传入面板所属的设备组件Item对象，以便从Item中加载相应的事件响应方法
        :return:
        """
        if tab_name == "SpecificationValues" or tab_name == "规格值":
            tab_items = device_item.SpecificationValues or {}
            panel_info_tab, connect_slot = self._form_display_panel(tab_items, device_item)
            if len(self.ui.sfv.children()) == 1 and isinstance(self.ui.sfv.children()[0], QLayout):
                self.ui.sfv.layout().addWidget(panel_info_tab)
            else:
                self.ui.sfv.layout().removeWidget(self.ui.sfv.children()[1])
                if len(self.ui.sfv.children()) > 1:
                    # 必须强制删除一次原组件，才能彻底删除干净，replaceWidget方法也存在删除不干净的问题
                    self.ui.sfv.children()[1].deleteLater()
                self.ui.sfv.layout().addWidget(panel_info_tab)

            values = device_item.values
            _values = values.get_initial(need_unit=True)
            panel_info_tab.set_values(_values)  # 此时，combobox连接事件尚未添加，因此设置值不会触发相应事件

            # 数值设置完成后，再添加combox链接事件
            # 如果在值设置之前连接事件，则会导致多个组件切换时，动态响应参数与响应时间源不匹配的问题
            self._add_combo_box_connect(panel_info_tab, connect_slot, device_item)

            return panel_info_tab
        elif tab_name == "Results" or tab_name == "结果":
            tab_items = device_item.Result or {}
            panel_info_tab, connect_slot = self._form_display_panel(tab_items, device_item)

            if len(self.ui.res.children()) == 1 and isinstance(self.ui.res.children()[0], QLayout):
                self.ui.res.layout().addWidget(panel_info_tab)
            else:
                self.ui.res.layout().removeWidget(self.ui.res.children()[1])
                if len(self.ui.res.children()) > 1:
                    # 必须强制删除一次原组件，才能彻底删除干净，replaceWidget方法也存在删除不干净的问题
                    self.ui.res.children()[1].deleteLater()
                self.ui.res.layout().addWidget(panel_info_tab)

            values = device_item.values
            _values = values.get_results(need_unit=True)
            panel_info_tab.set_values(_values)
        elif tab_name == "材料" or tab_name == "Materials":
            pass

    def _add_combo_box_connect(self, panel_info_tab, connect_slot, device_item):
        # 处理信息面板中combobox值改变时的事件响应
        if len(connect_slot) > 0:
            _ = []
            for connect in connect_slot:
                exist = False
                # 首先合并connect_slot中var和connect相同的信号与槽
                for c in _:
                    if c["var"] == connect["var"] and c["connect"] == connect["connect"]:
                        if isinstance(c["object"], list):
                            c["object"].append(connect["object"])
                            c["object_type"].append(connect["object_type"])
                        else:
                            c["object"] = [c["object"], connect["object"]]
                            c["object_type"] = [c["object_type"], connect["object_type"]]
                        exist = True
                        break
                if not exist:
                    _.append(connect)

            for connect in _:
                connect_func = eval(f"device_item.{connect.get('connect')}")
                var = connect.get("var")
                var_item = panel_info_tab.get_child_yk_items(var)
                value_combobox = var_item.value
                if isinstance(value_combobox, QComboBox):
                    # 多个connect_func方法相同时，则连接的函数传入的参数会相同，这会导致相当于信号触发最后一次绑定的事件多次，而不是各种参数
                    # 各调用一次，需要进行处理
                    value_combobox.currentTextChanged.connect(
                        lambda: self.combox_text_changed(connect_func, connect, panel_info_tab))

    def _form_display_panel(self, tab_items, device_item):
        """
        生成详细信息面板

        :param tab_items: 面板需要显示的数据，一般为item.SpecificationValues或item.Result数据对象
        :return:
        """
        panel_sfv = DisplayPanel()
        connect_slot = []  # 信息面板中各项之间可能有combobox等事件联动，记录下来
        for 参数类别, 参数项 in tab_items.items():
            disabled = False
            panel_参数类别 = QWidget()
            panel_参数类别.setLayout(QVBoxLayout())
            panel_参数类别.layout().setContentsMargins(0, 0, 0, 0)
            panel_参数类别.layout().setSpacing(1)
            for 子参数项 in 参数项:
                size = [100, 80, 60]
                name = 子参数项[0]
                symbol = 子参数项[1]
                value = 子参数项[2]
                if len(子参数项) >= 4:
                    unit = 子参数项[3]
                else:
                    unit = None
                if isinstance(unit, dict):
                    func = unit.get("func")
                    _res = re.findall(r"(.+)[(](.*)[)]", func)[0]
                    func_name, var = _res
                    # symbol的unit需要根据func_name(var)的执行结果确定
                    _slot = {"var": var, "connect": func_name, "object": symbol, "object_type": "unit"}
                    connect_slot.append(_slot)
                    try:
                        unit = eval(f"device_item.{func_name}('{var}')")
                        unit = {"value": unit.allowed_values, "selected_val": unit.unit}
                    except:
                        unit = ["初始化时发生错误！"]
                else:
                    pass
                if len(子参数项) >= 5:
                    var4 = 子参数项[4]
                    if isinstance(var4, str):
                        disabled = True if var4 == "disabled" else False
                    elif isinstance(var4, dict):
                        if var4.get("size") is not None:
                            size = var4.get("size")
                item = YkItem(f"{name}[{symbol}]", value, unit, size=size, draggable=True)
                # if isinstance(item.value, QLineEdit):
                #     item.value.setDragEnabled(True)
                if disabled:
                    item.value.setEnabled(False)
                panel_参数类别.layout().addWidget(item)
            panel_fold = YkFoldableWidget(参数类别, panel_参数类别)
            panel_sfv.layout().addWidget(panel_fold)

        return panel_sfv, connect_slot

    def combox_text_changed(self, func, connect_info, panel_sfv):
        """
        组件信息面板上，combobox组件值改变时，触发的事件处理

        :param func:
        :param connect_info:
        :return:
        """
        # noinspection all
        var, func = connect_info.get("var"), func  # 这两个参数在eval语句中使用
        self.set_values_panel2data()  # 先将当前改变后的值写入self.values中，再调用dynamic_unit方法
        res: Unit = eval(f"func(var)")
        object_yk_items = connect_info.get("object")
        object_types = connect_info.get("object_type")
        if not isinstance(object_yk_items, list):
            object_yk_items = [object_yk_items]
            object_types = [object_types]

        for object_item, object_type in zip(object_yk_items, object_types):
            des = panel_sfv.get_child_yk_items(object_item)  # 设置的目标YkItem对象
            if object_type == "unit":
                des.unit.setModel(QStringListModel(res.allowed_values))
                des.set_unit(res.unit)

    def set_values_panel2data(self):
        """
        将面板上的数值赋值给组件的values对象，点击【确定】按钮时会执行该方法
        """
        # 获取到当前正在设置的组件，根据场景id
        _ = self.ui.detail_tab.tabText(0).split("/ S")
        if len(_) == 1:
            return
        scene_id = int(_[1].replace(']', '').strip())
        item = self.content_widget.scene.items_scene_id.get(scene_id)
        if item is None:  # 说明当前的组件已经被删除，因为显示下一个信息面板时会首先保存上一个信息，所以该方法可能在组件已删除的情况下被调用
            return
        # 获取当前设置的面板名
        tab_name = self.ui.info_tab.tabText(self.ui.info_tab.currentIndex())
        if tab_name == "规格值" or tab_name == "SpecificationValues":
            _layout = self.ui.sfv.layout()  # self.ui.sfv.layout()没有子项
            if len(self.ui.sfv.children()) > 1 and isinstance(self.ui.sfv.children()[1], DisplayPanel):
                panel: DisplayPanel = self.ui.sfv.children()[1]  # 当前sfv上显示的面板
                values: dict = panel.get_values()  # 获取面板上的所有值，包括空值
                item.values.set_values(0, values, set_none=True)  # 将面板上的值赋值给设备组件，包括空值，面板上获取的值都是初始状态的值
                item.after_values_updated()  # 值设置完成后进行的操作
        elif tab_name == "结果":
            pass  # 结果不保存至当前设备组件，结果数据只能是计算出来的

    def set_values_data2panel(self):
        """
        将组件中的参数数值显示到面板上
        """
        # 获取到当前正在设置的组件，根据场景id
        scene_id = int(self.ui.detail_tab.tabText(0).split("/ S")[1].replace(']', '').strip())
        item = self.content_widget.scene.items_scene_id.get(scene_id)

        # 获取当前设置的面板名
        tab_name = self.ui.info_tab.tabText(self.ui.info_tab.currentIndex())
        if tab_name == "规格值" or tab_name == "SpecificationValues":
            panel: DisplayPanel = item.SpecificationValuesPanel
            children = self.ui.sfv.children()
            self.ui.sfv.layout().replaceWidget(children[1], panel)
            values = panel.set_values(item.values)

    def cancel(self):
        ...

    def info_tab_changed(self, idx):
        """


        :param idx:
        :return:
        """
        # 获取到当前正在设置的组件，根据场景id
        _ = self.ui.detail_tab.tabText(0).split("/ S")
        if len(_) == 1:
            return
        scene_id = int(_[1].replace(']', '').strip())
        # 获取到当前的信息面板名，如 "规格值"、"结果"
        tab_name = self.ui.info_tab.tabText(idx)
        self.display_info(scene_id=scene_id, tab=tab_name, save_data=False)  # 切换面板触发的显示信息不保存数据

    def get_color(self):
        color = QColorDialog.getColor(Qt.white)
        if color.isValid():
            self.content_widget.scene.setBackgroundBrush(QBrush(color))

    def clear_log_console(self):
        log_editor: YkEditor = self.ui.log_editor
        log_editor.setText("")

    def get_calculate_info(self):
        """
        获取计算设置信息
        """
        settings = self.proj.get('settings')
        if settings is None:
            return None, (None,)
        else:
            cal_type = settings.get('cal_type')
            cal_time_interval = settings.get('cal_time_interval')
        return cal_type, (cal_time_interval,)

    def set_calculate(self):
        """
        计算设置，点击后默认10s计算一次
        {"settings":
            {
                "cal_type": "real_time",
                "cal_time_interval": 10
            }
        }
        """
        settings = self.proj.get('settings')
        if settings is None:
            self.proj.update({"settings": {"cal_type": "real_time", "cal_time_interval": 10}})
        else:
            if settings.get("cal_type") == "real_time":
                self.proj.update({"settings": {"cal_type": "normal"}})
            else:
                self.proj.update({"settings": {"cal_type": "real_time", "cal_time_interval": 10}})


if __name__ == "__main__":
    run_app(MainWindow)



# 量化分析中使用的数据库对应的类
from sqlalchemy import Column, String, DATETIME, BOOLEAN, DATE, Date

from yangke.dataset.sqlalchemy import Base


class Holiday(Base):
    __tablename__ = "Holiday"
    calendarDate = Column(Date, nullable=False, primary_key=True)
    isOpen = Column(BOOLEAN, nullable=False)

    def __init__(self, date, is_open):
        self.calendarDate = date
        self.isOpen = is_open

    def __repr__(self):
        return f"Holiday({self.calendarDate}: {self.isOpen})"



# encoding=utf8
"""
制作*.whl的安装文件
"""
import sys
import os

__version__ = "1.15.7"
extras_require = {  # 额外依赖项，安装方法“pip install yangke[Database]”
    "Database": [
        'pymysql>=0.9.3',
        'DBUtils>=1.3',  # 用于创建数据库连接池
        'cryptography>=3.1.1',  # 用于mysql8以上版本的连接加密，必须安装，否则无法解析密码
        'python-docx',
        'mysql-connector-python>=8.0.24',
        'sqlalchemy<=1.4.39',  # 2.0版本的sqlalchemy存在很多兼容性问题，还需要进一步解决
    ],
    "windows": [
        'pypiwin32',
    ],
    "web": [
        'flask>=1.1.2',
        'Flask-Cors>=3.0.8',
        'waitress',
    ],
    "Stock": [
        # 'torch>=1.4.0', # pytorch需要单独安装，pypi里版本太老
        'tushare>=1.2.48',
        'scrapy>=2.0.0',
        'sxtwl>=1.0.7',
        'pymysql>=0.9.3',
        'DBUtils>=1.3',
        'selenium>=3.141.0',
        'mysql-connector-python>=8.0.24',
        'selenium>=3.141.0',
        'sqlalchemy>=1.3.23',
    ],
    "ImageRecognition": [
        'opencv-python>=4.2.0',
        'cmake>=3.18.2',
        'boost',
        # 'dlib>=19.17.0',  # dlib需要单独安装，涉及到cmake和boost
        'pillow>=7.0.0',

        'requests>=2.22.0',
        'tensorflow>=2.8.0',
        'torch>=1.11.0',
        'optuna',
        'plotly',
        'paddlepaddle',
        'paddleocr>=2.0.1',

    ],
    "GameServer": [
        'twisted>=20.3.0',
        'flask>=1.1.2',
        'flask_cors>=3.0.8',
        'requests>=2.22.0',
        'gevent>=20.5.0',
        'gevent-websocket',
        'waitress>=1.4.4',
        'lxml',
    ],
    "Performance": [
        'iapws>=1.5.2',
        'pygame',
        # 'PyQt5',
        # 安装PyQt5时如果出现ERROR: Could not install packages due to an OSError: [Errno 13] Permission denied:
        # ‘C:\Users\zhangbin\AppData\Roaming\Python\Python39\site-packages\PyQt5\Qt5\bin\d3dcompiler_47.dll’，
        # 说明d3dcompiler_47.dll文件占用，禁止无关应用开机启动，重启安装尝试。
        # 'PyQtWebEngine',  # 用于显示pyecharts图形
        'PyQt6',
        'PyQt6-QScintilla'
        # 'PySide6'
    ],

}

module_name_list = list(extras_require.keys())


def info():
    from yangke.common.config import printInColor
    printInColor(" module 'yangke' installed successfully ", color_fg='white', color_bg='cyan', mode=1)
    printInColor(" version is {} ".format(__version__), color_fg="white", color_bg="yellow", mode=1)
    print("The optional submodules are: ", end="")

    for mod in module_name_list[:-1]:
        printInColor('[{}]'.format(mod), mode=1, color_bg='', end='')
        print(", ", end='')
    printInColor('[{}]'.format(module_name_list[-1]), color_bg='', mode=1)

    print("Use command ", end="")  # end设置不换行输出
    printInColor("pip install yangke", color_fg='red', color_bg='yellow', end='', mode=1)
    printInColor("[Database]", color_bg='yellow', end="", mode=1)
    print("/", end='')
    printInColor("pip install *.whl", color_fg='red', color_bg='yellow', end='', mode=1)
    printInColor("[Database]", color_bg='yellow', end="", mode=1)
    print(" to install the selected submodule.")

    print("Use command ", end="")
    printInColor("pip install *.whl[All]", color_fg='red', color_bg='yellow', end='', mode=1)
    printInColor("[Database]", color_bg='yellow', end="", mode=1)
    print(" to install all the submodule.")


def yangke_test():
    info()


def version():
    info()


def test():
    info()

# yangke_test()



import math
from enum import Enum, unique

import pandas as pd

from yangke.base import get_key_value, is_number, interpolate_value_complex, interpolate_value_simple
from yangke.common.config import logger
from yangke.common.fileOperate import read_excel


@unique
@get_key_value
class TagsRead(Enum):
    机组1循环水泵1电流 = "DCS1:PAC20AP001CE"
    机组1循环水泵2高速电流 = "DCS1:PAC20AP002HCE"
    机组1循环水泵2低速电流 = "DCS1:PAC20AP002LCE"
    机组1循环水泵3高速电流 = "DCS1:PAC20AP003HCE"
    机组1循环水泵3低速电流 = "DCS1:PAC20AP003LCE"
    机组1循环水泵4电流 = "DCS1:PAC20AP004CE"

    机组1高压凝汽器压力1 = "DCS1:MAG10CP101"
    机组1高压凝汽器压力2 = "DCS1:MAG10CP102"
    机组1高压凝汽器压力3 = "DCS1:MAG10CP103"
    机组1高压凝汽器压力4 = "DCS1:MAG10CP104"
    机组1低压凝汽器压力1 = "DCS1:MAG20CP101"
    机组1低压凝汽器压力2 = "DCS1:MAG20CP102"
    机组1低压凝汽器压力3 = "DCS1:MAG20CP103"
    机组1低压凝汽器压力4 = "DCS1:MAG20CP104"

    机组1凝汽器真空 = "DCS1:DMNQQZK"
    机组1低压凝汽器循环水进水温度2 = "DCS1:PAB33CT301"
    机组1低压凝汽器循环水进水温度1 = "DCS1:PAB31CT301"
    机组1高压凝汽器循环水出水温度2 = "DCS1:PAB34CT301"
    # 机组1高压凝汽器循环水出水温度1 = "DCS1:PAB32CT302"
    机组1低压凝汽器进水压力 = "DCS1:PAB33CP101"
    机组1高压凝汽器出水压力 = "DCS1:PAB32CP101"
    机组1大气压力1 = "DCS1:PAD00CP101"
    机组1大气压力2 = "DCS1:PAD00CP102"
    # 机组1大气压力3 = "DCS1:PAC00CP103"
    机组1环境温度1 = "DCS1:PAD00CT311"
    机组1环境温度2 = "DCS1:PAD00CT312"
    机组1环境温度3 = "DCS1:PAD00CT313"
    机组1环境温度4 = "DCS1:PAD00CT314"
    机组1环境温度5 = "DCS1:PAD00CT315"
    机组1环境温度6 = "DCS1:PAD00CT316"
    机组1环境温度7 = "DCS1:PAD00CT317"
    机组1环境温度8 = "DCS1:PAD00CT318"
    机组1环境温度9 = "DCS1:PAD00CT319"
    发电机功率1 = "DCS1:LOAD"
    总燃料量1 = "DCS1:TFFREAL"

    机组2循环水泵1电流 = "DCS2:PAC20AP001CE"
    机组2循环水泵2高速电流 = "DCS2:PAC20AP002HCE"
    机组2循环水泵2低速电流 = "DCS2:PAC20AP002LCE"
    机组2循环水泵3高速电流 = "DCS2:PAC20AP003HCE"
    机组2循环水泵3低速电流 = "DCS2:PAC20AP003LCE"
    机组2循环水泵4电流 = "DCS2:PAC20AP004CE"

    机组2高压凝汽器压力1 = "DCS2:MAG10CP101"
    机组2高压凝汽器压力2 = "DCS2:MAG10CP102"
    机组2高压凝汽器压力3 = "DCS2:MAG10CP103"
    机组2高压凝汽器压力4 = "DCS2:MAG10CP104"
    机组2低压凝汽器压力1 = "DCS2:MAG20CP101"
    机组2低压凝汽器压力2 = "DCS2:MAG20CP102"
    机组2低压凝汽器压力3 = "DCS2:MAG20CP103"
    机组2低压凝汽器压力4 = "DCS2:MAG20CP104"

    机组2凝汽器真空 = "DCS2:DMNQQZK"
    机组2低压凝汽器循环水进水温度2 = "DCS2:PAB33CT301"
    机组2低压凝汽器循环水进水温度1 = "DCS2:PAB31CT301"
    机组2高压凝汽器循环水出水温度2 = "DCS2:PAB34CT301"
    # 机组2高压凝汽器循环水出水温度1 = "DCS2:PAB32CT302"
    机组2低压凝汽器进水压力 = "DCS2:PAB33CP101"
    机组2高压凝汽器出水压力 = "DCS2:PAB32CP101"

    机组2大气压力1 = "DCS2:PAD00CP101"
    机组2大气压力2 = "DCS2:PAD00CP102"
    # 机组2大气压力3 = "DCS2:PAC00CP103"
    机组2环境温度1 = "DCS2:PAD00CT311"
    机组2环境温度2 = "DCS2:PAD00CT312"
    机组2环境温度3 = "DCS2:PAD00CT313"
    机组2环境温度4 = "DCS2:PAD00CT314"
    机组2环境温度5 = "DCS2:PAD00CT315"
    机组2环境温度6 = "DCS2:PAD00CT316"
    机组2环境温度7 = "DCS2:PAD00CT317"
    机组2环境温度8 = "DCS2:PAD00CT318"
    机组2环境温度9 = "DCS2:PAD00CT319"
    发电机功率2 = "DCS2:LOAD"
    总燃料量2 = "DCS2:TFFREAL"


@unique
@get_key_value
class TagsWrite(Enum):
    建议循泵运行方式11 = "DCS1:P_NUM_OPT1"
    建议循泵运行方式12 = "DCS1:P_NUM_OPT2"
    建议循泵运行方式13 = "DCS1:P_NUM_OPT3"
    建议循泵运行方式14 = "DCS1:P_NUM_OPT4"
    建议循泵运行方式21 = "DCS2:P_NUM_OPT1"
    建议循泵运行方式22 = "DCS2:P_NUM_OPT2"
    建议循泵运行方式23 = "DCS2:P_NUM_OPT3"
    建议循泵运行方式24 = "DCS2:P_NUM_OPT4"
    最优运行方式下背压1 = "DCS1:BP_Best_Con"
    最优运行方式下背压2 = "DCS2:BP_Best_Con"
    优化后机组节煤量1 = "DCS1:Coal_Saving_Run"
    优化后机组节煤量2 = "DCS2:Coal_Saving_Run"
    循环水流量1 = "DCS1:Cycle_F"
    循环水流量2 = "DCS2:Cycle_F"
    循泵总功率1 = "DCS1:Power_Pump"
    循泵总功率2 = "DCS2:Power_Pump"
    净电功率增益1 = "DCS1:Power_Net_Inc"
    净电功率增益2 = "DCS2:Power_Net_Inc"

    循泵切换小时数11 = "Dcs1:xhsbtime_hour"
    循泵切换分钟数11 = "Dcs1:xhsbtime_minute"
    循泵切换小时数21 = "Dcs1:xhsbtime_hour2"
    循泵切换分钟数21 = "Dcs1:xhsbtime_minute2"
    循泵切换小时数31 = "Dcs1:xhsbtime_hour3"
    循泵切换分钟数31 = "Dcs1:xhsbtime_minute3"
    循泵切换小时数41 = "Dcs1:xhsbtime_hour4"
    循泵切换分钟数41 = "Dcs1:xhsbtime_minute4"
    循泵切换小时数12 = "Dcs2:xhsbtime_hour"
    循泵切换分钟数12 = "Dcs2:xhsbtime_minute"
    循泵切换小时数22 = "Dcs2:xhsbtime_hour2"
    循泵切换分钟数22 = "Dcs2:xhsbtime_minute2"
    循泵切换小时数32 = "Dcs2:xhsbtime_hour3"
    循泵切换分钟数32 = "Dcs2:xhsbtime_minute3"
    循泵切换小时数42 = "Dcs2:xhsbtime_hour4"
    循泵切换分钟数42 = "Dcs2:xhsbtime_minute4"

    高速泵运行台数实时值1 = "DCS1:NUM_CWP_H"
    低速泵运行台数实时值1 = "DCS1:NUM_CWP_L"
    高速泵运行台数推荐值1 = "DCS1:NUM_CWP_H_OPT"
    低速泵运行台数推荐值1 = "DCS1:NUM_CWP_L_OPT"
    高速泵运行台数实时值2 = "DCS2:NUM_CWP_H"
    低速泵运行台数实时值2 = "DCS2:NUM_CWP_L"
    高速泵运行台数推荐值2 = "DCS2:NUM_CWP_H_OPT"
    低速泵运行台数推荐值2 = "DCS2:NUM_CWP_L_OPT"
    循泵总功率实际值1 = "DCS1:Power_Pump_Act"
    循泵总功率实际值2 = "DCS2:Power_Pump_Act"

    阻塞背压1 = "DCS1:BP_ZUSE"
    阻塞背压2 = "DCS2:BP_ZUSE"


def load_data(
        file=r"C:\Users\YangKe\Documents\WPS Cloud Files\217145378\6科研项目\2022\榆能横山\冷端优化软件输入数据.xlsx"):
    """
    加载性能试验优化结果

    :param file:
    :return:
    """
    backpressure = {}
    dfs = read_excel(file=file, need_merge=False, index_col=0)
    for sheet in ["900MW", "800MW", "700MW", "600MW", "500MW", "400MW"]:
        df = dfs[sheet]
        df.dropna(how="all", axis=0, inplace=True)
        backpressure.update({sheet: df})
    type_opt = dfs["最佳运行方式"]
    pump_info = dfs["循泵"]
    coal_info = dfs["煤耗"]
    return backpressure, type_opt, pump_info, coal_info


def get_backpressure(power=450, type_="四台高速", t=23, bp_total=None):
    """
    根据机组负荷、循泵运行方式、循环水温度计算凝汽器背压

    :param power:
    :param type_:
    :param t:
    :return:
    """
    lower_power = math.floor(power / 100) * 100
    upper_power = math.ceil(power / 100) * 100
    if lower_power == 300:  # 小于300MW时，用400和500MW的数据插值计算
        lower_power = 500
    if upper_power == 1100:
        upper_power = 900
    lower_power_idx = str(f"{lower_power}MW")
    upper_power_idx = str(f"{upper_power}MW")
    lower_df = bp_total[lower_power_idx]
    upper_df = bp_total[upper_power_idx]
    if type_ in ["一台高速", "一台低速"]:
        type_ = "两台低速"
    lower_type = lower_df[type_]
    upper_type = upper_df[type_]

    lower_pressure = interpolate_value_complex(t, x_list=[float(i.replace("℃", "")) for i in lower_type.keys()],
                                               y_list=lower_type)
    upper_pressure = interpolate_value_complex(t, x_list=[float(i.replace("℃", "")) for i in upper_type.keys()],
                                               y_list=upper_type)
    pressure = interpolate_value_simple(power, x1=lower_power, y1=lower_pressure, x2=upper_power, y2=upper_pressure)
    return pressure


def get_opt_type(power=460, t=22.5, type_opt=None):
    power_ = round(power / 100) * 100
    power_idx = f"{power_}MW"
    t_ = round(t / 2) * 2
    t_idx = f"{t_}℃"
    return type_opt.loc[t_idx, power_idx]


def get_flowrate(type_, pump_info):
    if type_ not in ["四台高速", "三高一低", "两高两低", "三台高速", "两高一低", "一高两低", "两台高速", "一高一低",
                     "两台低速"]:
        logger.warning(f"查询的循泵运行方式不存在：{type_}")
        return 0
    return pump_info.loc[type_, "循环循环水流量(t/h)"]


def get_pump_power(type_, pump_info):
    if type_ not in ["四台高速", "三高一低", "两高两低", "三台高速", "两高一低", "一高两低", "两台高速", "一高一低",
                     "两台低速"]:
        logger.warning(f"查询的循泵运行方式不存在：{type_}")
        return 0
    return pump_info.loc[type_, "循环水泵功率(kW)"]


def get_coal_consume(power, coal_info):
    """
    计算发电机功率对应的煤耗，功率可以超出coal_info给定的范围，外延插值

    :param power:
    :param coal_info:
    :return:
    """
    x_list = list(coal_info.index)
    y_list = list(coal_info["供电煤耗g/(kW·h)"])
    return interpolate_value_complex(power, x_list=x_list, y_list=y_list)


def get_dpower_per_kpa(power, coal_info):
    """
    计算背压变化1kPa引起的发电机处理变化，功率可以超出coal_info给定的范围，外延插值

    :param power:
    :param coal_info:
    :return:
    """
    x_list = list(coal_info.index)
    y_list = list(coal_info["背压1kPa出力增加（kW）"])
    return interpolate_value_complex(power, x_list=x_list, y_list=y_list)



import datetime
import traceback
from typing import Any

import pandas as pd
import sqlalchemy.sql.dml
from sqlalchemy import create_engine, MetaData, Column, inspect, Table, String, text, insert, delete, update, select
from sqlalchemy.dialects.mysql import INTEGER, DOUBLE, BIGINT, VARCHAR, CHAR, TEXT, DATETIME
from sqlalchemy.engine import Engine, Row
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import sessionmaker, Query
from sqlalchemy.sql import Select

from yangke.common.config import logger

# # 创建基类, 使用ORM方式操作数据库是继承该类
Base = declarative_base()
# noinspection all
修改时间记录表 = 'modify_time'


class YkTable(Table):
    def __init__(self, name, metadata, columns=None):
        if columns is None:
            columns = []
        super().__init__(name, metadata, *columns)


class YkColumn(Column):
    def __init__(self, name, dtype, primary_key=False, nullable=False):
        """
        Sqlite数据库不只是foreignKey关键字
        """
        super().__init__(name, dtype, primary_key=primary_key, nullable=nullable)


# class Base(Base_):
#     __tablename__ = "default"  # 也可以是 __tablename__
#
#     def __init__(self):
#         # super().__init__()
#         ...


class ModifyTime(Base):
    """
    用来记录数据库表格最后更新时间的表格，因为目前MySql和Sqlite数据库均无法查询表格的最后更新时间，因此使用数据库表格记录所有表格的最后更新
    时间
    """
    __tablename__ = 修改时间记录表
    table = Column(String(50), nullable=False, primary_key=True)  # autoincrement=True
    datetime = Column(DATETIME, nullable=False)

    def __init__(self, table, date_time):
        self.table = table
        self.datetime = date_time

    def __repr__(self):
        return f"ModifyTime(table={self.table}, 修改时间={self.datetime})"


class SqlOperator:
    def __init__(self, engine: Engine = None):
        """
        SQL数据库操作类，同时支持MySql和Sqlite数据库，使用示例：
        engine = create_engine("mysql+mysqlconnector://root:password@localhost:3306", pool_recycle=7200)
        engine = create_engine('sqlite:///stocks.db', echo=True)
        so = SqlOperator(engine=engine)
        然后可以调用so的增删改查等方法，所有方法均同时支持MySql和Sqlite数据库

        可能遇到的问题：
        1.sqlalchemy.exc.InterfaceError: (mysql.connector.errors.InterfaceError) 2003: Can't connect to MySQL server on
        'disk.yangke.site:3306' (10060 由于连接方在一段时间后没有正确答复或连接的主机没有反应，连接尝试失败。)
        解决方法： 不使用mysqlconnector连接，pip install mysql后使用：create_engine("mysql://root:password@localhost:3306")
        """
        # 创建数据库引擎，ps:这里并没有连接具体数据库
        # engine = create_engine("mysql+mysqlconnector://root:password@localhost:3306/db?charset=utf8", pool_recycle=7200)

        # 连接时，如果报Character set 'utf8' unsupported错误，且数据库文件中没有非ascii字符，可以尝试切换charset的值。
        # 已知charset可取值包括(ascii, utf8, utf8mb4, gbk, cp1250)，可以执行show collation语句查看mysql数据库支持的字符集。

        self.engine = engine
        self.insp = inspect(self.engine)
        self.meta_data = MetaData()  # 兼容sqlalchemy 2.0版本的写法，bind参数写在reflect()方法中
        self.meta_data.reflect(bind=engine)
        self.session = sessionmaker(bind=engine)()  # 使用sessionmaker保持数据库会话连接
        self.base = Base
        self.connect = self.engine.connect()

    def create_table(self, table_name, columns):
        """
        创建数据库表，如果同名表存在，会报错
        示例1：
        self.create_table(table_name=f"daily{symbol}",
                          columns=[
                              YkColumn('trade_date', DATE(), primary_key=True),
                              YkColumn('open', Float()),
                              YkColumn('high', Float()),
                              YkColumn('low', Float()),
                              YkColumn('close', Float()),
                              YkColumn('vol', Float()),
                              YkColumn('amount', Float()),
                          ])
        """
        table = Table(table_name, self.meta_data, *columns)
        logger.debug(f"创建表格")
        table.create(bind=self.connect)  # bind=self.engine时，部分情况下会卡住

    def create_all_base_table(self):
        """
        创建所有继承自Base类的Python类的映射数据库表。也就是说，只要定义了继承自本模块中Base类的类，则该类会被映射成一个数据库表，
        本方法会自动创建所有映射的数据库表。
        """
        self.base.metadata.create_all(self.engine)

    def get_type_of_column(self, table_name=None, column_name=None):
        """
        获取mysql表中字段的类型，如果不设置column_name则返回所有的字段类型
        :param table_name:
        :param column_name: 为空则依次返回所有列的类型，封装为一个列表
        :return:
        """
        cols = self.insp.get_columns(table_name)
        res = None
        if column_name is None:
            res = {}
            for col in cols:
                res.update({col["name"]: col["type"]})
        else:
            for col in cols:
                if col["name"] == column_name:
                    res = col["type"]
                    break
        return res

    def get_column_names(self, table_name):
        """
        获取表格中的列名，返回列名列表
        """
        cols = self.insp.get_columns(table_name)
        res = [col["name"] for col in cols]
        return res

    def get_update_time_of_table(self, table_name):
        """
        获取表的最后更新时间
        """
        if self.has_table(table_name):
            # 查询modifyTime表，modifyTime表中记录了所有表格的最后更新时间
            if self.has_table(修改时间记录表):
                table: Table = self.get_table(修改时间记录表)
                select: Select = table.select().where(table.c.table == table_name)  # 构建一个选择语句
                res = self.session.execute(select)  # 执行选择语句并获取结果
                res = res.fetchone()
                if res is None:
                    return None
                return res.datetime  # 改写法兼容sqlalchemy1.4和2.0
                # return res["datetime"]  # 改写发不兼容sqlalchemy2.0
            else:
                return None
        else:
            return None

    def update_update_time_of_table(self, table_name):
        """
        更新表的最后更新时间
        """
        if not self.has_table(修改时间记录表):
            self.create_all_base_table()

        now = datetime.datetime.now()
        if self.exists_in_table(table_name=修改时间记录表, col_name='table', value=table_name):
            # stmt = update(table).where(table.c.table == table_name).values(datetime=now)  # 兼容写法
            # self.connect.execute(stmt)
            self.update_item(table_name=修改时间记录表, conditions={"table": table_name}, values={"datetime": now})
        else:
            self.session.add(ModifyTime(table_name, now))
        self.session.commit()

    def exists_in_table(self, table_name: str = None, col_name: str = None, value: str = None,
                        condition_dict: dict = None,
                        return_result: bool = False):

        """
        表tableName中是否存在列col_name的值位value的行

        :param table_name: 表名
        :param col_name: 列名
        :param value: 列的值
        :param condition_dict: 查询的键值对字典值，优先于col_name和value传入的值，即会覆盖col_name和value传入的值
        :param return_result: 是否需要返回查找到的数据行，如果为真，则返回所有符合查找条件的数据行
        :return:
        """
        if return_result:
            first_or_all = 'all'
        else:
            first_or_all = 'first'

        condition_dict = condition_dict or {}
        if col_name is not None and value is not None:
            condition_dict.update({col_name: value})

        res = self.select_item(table_name, condition_dict, first_or_all=first_or_all)
        if res is not None:
            return True
        else:
            return False

    def select_in_table(self, table_name=None, condition_dict: dict = None, result_col: list | str = None, limit=10,
                        offset=0,
                        fuzzy=False, first_or_all="first", result_type=None, cls=None, **kwargs) -> Row | list | Any:
        """
        查

        精确查询，设置fuzzy为True，且condition_dict中的value为字符串值
        模糊查询，设置fuzzy为False，日期列不能使用模糊查询，只能使用范围查询
        范围查询，设置fuzzy为True，且condition_dict中的value为长度为2的列表，列表第一、二项分别为范围下、上限，且包含上下限

        当result_type=="json"时，返回的是一个list(dict)的json对象，即[{col1: value1, col2: value2,...}, ...}的json对象
        列表的每一项对应一条匹配的查询结果
        每一项的字典分别是{列名：值}

        kwargs={"date_format": "%Y-%m-%d %H:%M:%S"} 如果mysql中存在日期列，需要将日期转换为字符串，该参数定义日期字符串格式

        示例1：
        fetch = self.sql.select_in_table(cls=Holiday, condition_dict={"calendarDate": day_datetime},
                                         result_col=['isOpen'])

        :param table_name: 当使用传统查询方式时，需要传入数据库表名
        :param cls: 当使用ORM模型时，只需要传入数据库表在python中对应的映射类名，如果通过该方法查询，则查询到的数据会被自动转换为cls对象
        :param condition_dict:
        :param result_col: 不传入或传入空列表，则返回数据库中所有列
        :param limit:
        :param offset:
        :param fuzzy: 是否模糊查询
        :param first_or_all: 返回满足条件的第一个还是所有，支持"first", "all",
        :param result_type: 返回类型，如果为json，则返回为json格式的字符串
        :return: None或查询的列值列表或数据条的列表或sqlalchemy.engine.Row对象，出错时返回None，如列明不存在等；否则返回一个tuple类型的数据，长度为0表示未查询到满足条件的数据
        """
        # ---------------------- 如果使用table_name查询，则构建Table对象 ------------------------
        # Table对象可以当cls一样使用
        if cls is None:
            if self.has_table(table_name):
                table: Table = self.get_table(table_name)
                if table is None:
                    return None
                cls = table
        # ---------------------- 如果使用table_name查询，则构建Table对象 ------------------------
        # ----------------------- 根据条件字典，构建查询条件 --------------------------------
        if condition_dict is None:  # 查询表格中某列的所有值
            if result_col is None:
                return
            elif isinstance(result_col, str):
                res = self.session.execute(select(cls.c.get(result_col))).fetchall()
                res = [i._data[0] for i in res]
            else:  # list
                n = len(result_col)
                _1 = []
                for i in range(n):
                    _1.append(f"cls.c.get(result_col[{i}])")
                _2 = ",".join(_1)

                _3 = f"select({_2})"
                try:
                    _4 = eval(_3)
                    res = self.session.execute(_4).fetchall()
                    res = [i._data for i in res]
                except:
                    traceback.print_exc()
                    logger.error(f"检查返回列的名称是否正确：{result_col=}")
                    res = None
            return res
        # condition = []
        # for col, val in condition_dict.items():
        #     if isinstance(val, datetime.datetime) or isinstance(val, datetime.date):
        #         condition.append(f"{col}={val.__repr__()}")
        #     else:
        #         condition.append(f"{col}='{val}'")
        # condition = ",".join(condition)
        # ----------------------- 根据条件字典，构建查询条件 --------------------------------

        # .filter_by(node=node, password=password).all()  # filter()不支持组合查询，filter_by支持
        # _: Query = self.session.query(cls)
        # expression = f"_.filter_by({condition})"
        # _: Query = eval(expression)
        # item = eval(f"_.{first_or_all}()")

        _1 = []
        vals = []
        i = 0
        for k, v in condition_dict.items():
            vals.append(v)  # 为了兼容v为datetime或date类型的参数
            _1.append(f"cls.c.{k}==vals[{i}]")
            i += 1
        _2 = ",".join(_1)

        _3 = f"select(cls).where({_2})"
        items = self.session.execute(eval(_3)).fetchall()
        if len(items) == 0 or items is None:
            return None
        if first_or_all == "first":
            item = items[0]
        else:
            item = items

        # ------------------------ 如果指定了返回的数据列，则取出数据列并返回 -----------------------
        if isinstance(item, str):
            item = [item]
        if item is None or len(item) == 0:
            return item
        else:
            try:
                if isinstance(item, Row):  # item是个sqlalchemy.engine.row.Row对象，如读取Holiday表时会出现该情况
                    if result_col is None:
                        return item
                    elif isinstance(result_col, str):
                        # sqlalchemy2.0中，item.__getattr__(result_col)语法是正确的
                        # sqlalchemy1.4中，item.__getitem__(result_col)语法是正确的
                        res = eval(f"item.{result_col}")  # 该写法兼容sqlalchemy2.0
                        return res
                    elif isinstance(result_col, list) and len(result_col) == 1:
                        res = eval(f"item.{result_col[0]}")  # 该写法兼容sqlalchemy2.0
                        return res
                # noinspection all
                _ = [getattr(item, col) for col in item]  # table_name和cls两种方法都适用
                if len(_) == 1:
                    _ = _[0]
                return _
            except AttributeError:
                logger.error(item)
        # ------------------------ 如果指定了返回的数据列，则取出数据列并返回 -----------------------

    def select_item(self, table_name=None, condition_dict: dict = None, result_col: list | str = None, limit=10,
                    offset=0, first_or_all="first",
                    fuzzy=False, result_type=None):
        """
        查

        精确查询，设置fuzzy为True，且condition_dict中的value为字符串值
        模糊查询，设置fuzzy为False，日期列不能使用模糊查询，只能使用范围查询
        范围查询，设置fuzzy为True，且condition_dict中的value为长度为2的列表，列表第一、二项分别为范围下、上限，且包含上下限

        当result_type=="json"时，返回的是一个list(dict)的json对象，即[{col1: value1, col2: value2,...}, ...}的json对象
        列表的每一项对应一条匹配的查询结果
        每一项的字典分别是{列名：值}

        kwargs={"date_format": "%Y-%m-%d %H:%M:%S"} 如果mysql中存在日期列，需要将日期转换为字符串，该参数定义日期字符串格式

        示例1：
        fetch = self.sql.select_in_table(cls=Holiday, condition_dict={"calendarDate": day_datetime},
                                         result_col=['isOpen'])

        :param table_name: 当使用传统查询方式时，需要传入数据库表名
        :param condition_dict:
        :param result_col: 不传入或传入空列表，则返回数据库中所有列
        :param limit:
        :param offset:
        :param fuzzy: 是否模糊查询
        :param first_or_all: 返回满足条件的第一个还是所有，支持"first", "all",
        :param result_type: 返回类型，如果为json，则返回为json格式的字符串
        :return: None或查询的列值列表或数据条的列表或sqlalchemy.engine.Row对象，出错时返回None，如列明不存在等；否则返回一个tuple类型的数据，长度为0表示未查询到满足条件的数据
        """
        return self.select_in_table(table_name=table_name, condition_dict=condition_dict, result_col=result_col,
                                    limit=limit, offset=offset,
                                    fuzzy=fuzzy, first_or_all=first_or_all, result_type=result_type)

    def update_item(self, table_name, conditions: dict, values):
        """
        self.update_item('user', {"id": 1}, values={"name": "Tom"})

        兼容sqlalchemy 1.4及2.0以上版本。
        query.filter()或query.filter_by()的写法无法同时兼容sqlalchemy 1.4和2.0版本

        Parameters
        ----------
        table_name
        conditions: 数据表的索引列的值，必须唯一确定某一个记录
        values: 数据表中其他列的值，可以只设置部分列的值

        Returns
        -------

        """
        table: Table = self.get_table(table_name)
        # _1 = self.session.query(table)
        update_obj = update(table)
        for col_title, col_value in conditions.items():
            update_obj = update_obj.where(table.c.get(col_title) == col_value)  # 兼容写法

        dict_str = "{"
        objs = []
        i = 0
        for title, value in values.items():
            # if isinstance(value, datetime):
            #     dict_str = f"{dict_str}table.c.{title}: value,"
            objs.append(value)
            dict_str = f"{dict_str} table.c.{title}: objs[{i}],"  # 这样的写法是为了解决value是datetime或date类型时的兼容写法
            i += 1
        dict_str += "}"
        values = eval(dict_str)
        update_obj = update_obj.values(values)
        self.session.execute(update_obj)
        self.session.commit()

    def get_primary_col(self, table_name):
        """
        获取表的主键

        Parameters
        ----------
        table_name

        Returns
        -------

        """
        table = self.get_table(table_name)
        if isinstance(table, Table):
            # noinspection all
            for col in table.columns:
                if col.primary_key:
                    return col
        return None

    def insert_dataframe(self, table_name, df: pd.DataFrame, if_exists="append", index=False, dtype=None):
        """
        将DataFrame对象写入数据库中。兼容mysql和sqlite。df对象可以包含数据库中不存在的列，该方法会忽略这些列。

        pandas的to_sql方法存在以下问题：1.表格存在时，如果使用replace，则新创建的表的主键会丢失；2.append时会报重复键错误；

        批量插入数据时，
        简单说明：同样添加10W行数据，插入时间比
        1，Session.add(obj)   使用时间：6.89754080772秒
        2，Session.add(obj)注意：手动添加了主键id   使用时间：4.09481811523秒
        3，bulk_save_objects([obj1, obj2])   使用时间：1.65821218491秒
        4，bulk_insert_mappings(DBmodel, [dict1, dict2])  使用时间： 0.466513156781秒
        5，SQLAlchemy_core(DBmodel.__table__insert(), [dict1, dict2]) 使用时间：0.21024107933秒
        6，直接执行execute(str_sql_insert)  直接执行sql插入语句 使用时间：0.137335062027秒
        ————————————————
        版权声明：本文为CSDN博主「DHogan」的原创文章，遵循CC 4.0 BY-SA版权协议，转载请附上原文出处链接及本声明。
        原文链接：https://blog.csdn.net/dongyouyuan/article/details/79236673

        Parameters
        ----------
        table_name
        df
        if_exists: 参数参加pandas.DataFrame().to_sql()方法，可取值append, replace, fail
        index
        dtype : 指定df中各列的数值类型

        Returns
        -------

        """
        if self.has_table(table_name):  # 逐条插入，只要表存在，则不要使用pandas的to_sql方法，因为to_sql方法会覆盖原表，导致丢失主键
            primary_col = self.get_primary_col(table_name)
            primary_key = primary_col.name
            items_exist = self.select_item(table_name, result_col=primary_key)  # 获取数据库中表现有的数据记录的唯一标记
            if isinstance(items_exist, list):
                table: Table = self.get_table(table_name)
                # ---------------------- 判断需要插入的和需要更新的数据 ---------------------------
                df = df.fillna(0)  # 有些股票的交易量为nan，例如600607的1993-12-23日的交易量为nan
                df['new'] = df.apply(func=lambda row: False if row[primary_key] in items_exist else True, axis=1)
                # 2.0版本需要将Timestamp转换为str类型，pandas中的字符串在插入数据库时，会由str强制转换成DATE类型
                if len(df) > 0:
                    row1: pd.Series = df.iloc[0]  # df中的第一个记录
                    for title, value in row1.items():
                        if isinstance(value, pd.Timestamp):
                            df[title] = df[title].astype(str)

                df_new = df[df['new']]
                df_old = df[df['new'] == 0]  # 已存在的数据，视情况忽略或更新，0==False
                # ---------------------- 判断需要插入的和需要更新的数据 ---------------------------

                # ---------------------- 判断是否需要执行插入操作 ---------------------------------
                if if_exists == "append":  # 存在的数据不操作，只添加不存在的数据
                    if len(df_new) == 0:  # 如果只是追加不存在的数据，且插入的数据全部存在，则无须执行插入操作
                        return
                # ---------------------- 判断是否需要执行插入操作 ---------------------------------

                cols_table = self.get_column_names(table_name)

                # 按数据库表的列顺序排列dataframe列，无论if_exists取replace还是append，都要插入新值
                values_new = [dict(zip(cols_table, item)) for _, item in df_new[cols_table].iterrows()]
                try:
                    self.session.execute(table.insert(), values_new)  # 该方法速度仅次于直接执行sql语句，尽量使用该方法

                    if if_exists == "replace":  # 如果存在的数据需要替换
                        values_old = [dict(zip(cols_table, item)) for _, item in df_old[cols_table].iterrows()]
                        # self.session.bulk_update_mappings(table, values_old)
                        self.session.execute(table.update(), values_old)
                except:
                    traceback.print_exc()
                    logger.debug(f"{table=}, {values_new=}")
            else:
                logger.warning("未处理的数据插入")
        else:
            # 插入大数据量是特别慢，尤其是数据库已有重复记录时
            logger.warning(f"df.to_sql方法创建的表{table_name}没有主键，建议先创建表后，再插入数据")
            df.to_sql(table_name, self.engine, if_exists=if_exists, index=index)  # 该方法设置的数据库表没有主键

    def insert_item(self, table_name: str = None, values: list = None,
                    col_names: list = None, ignore=False,
                    replace=False, filter_warning=None):
        """
        向数据库中插入数据，这里传入的列名不要用反引号括起来，增

        values中的数据类型需要与表中每列的数据类型一致，数据库和python中数据类型对应如下：
        SqlAlchemy          python
        DATETIME            datetime.datetime/datetime.date
        VARCHAR             str

        cols_names和values两个列表一一对应。

        :param table_name: 表名
        :param values:
        :param col_names: 列名，如果是插入带有auto_increment属性的表数据，则必须指定列名，否则就需要指定auto_increment属性的字段的值
        :param ignore: 当插入数据重复时，是否忽略
        :param replace: 当插入数据重复时，是否替换，ignore和replace不能同时为True
        :param filter_warning: 过滤警告信息，[1062, 1265, 1366]，分别对应["Duplicate", "Data truncated", "Incorrect integer value"]
        """
        _: Table = self.get_table(table_name)

        # ------------------ 该段语句在sqlite服务器上测试成功，但mysql5.6服务器上测试数据不更新 -------------------
        # if col_names is None:
        #     col_names = self.get_column_names(table_name)
        # paras = []
        # for col, val in zip(col_names, values):
        #     if isinstance(val, datetime.datetime) or isinstance(val, datetime.date):
        #         val = val.__repr__()
        #         paras.append(f"{col}={val}")
        #     else:
        #         paras.append(f"{col}='{val}'")
        # paras = ",".join(paras)
        # state = f"_.insert().values({paras})"
        # eval(state)
        # self.session.commit()
        # ------------------ 该段语句在sqlite服务器上测试成功，但mysql5.6服务器上测试数据不更新 -------------------

        # ------------------ mysql5.6以下语句测试成功 -------------------------------
        if replace:
            columns = ""
            self.insert_item()
        else:
            ins = _.insert(values=dict(zip(col_names, values)))
            self.connect.execute(ins)
            self.session.commit()

    def has_table(self, table_name):
        """
        判断数据库中是否存在某个表
        """
        # return self.engine.has_table(table_name) # 该写法不兼容sqlalchemy 2.0版本
        return self.engine.dialect.has_table(self.engine.connect(), table_name)  # 该写法兼容sqlalchemy1.4和2.0版本

    def exist_table(self, table_name):
        """
        同has_table()
        """
        return self.has_table(table_name)

    def get_table(self, table_name=None):
        """
        获取表名为table_name的表对象，返回的是 Table()对象。
        如果不传入table_name，则返回数据库中的所有表，返回的是Table()对象的列表。
        """
        tables = [i for i in self.meta_data.tables.values()]
        if table_name is not None:
            for table in tables:
                if table.name == table_name:
                    return table
            return None
        return [i for i in self.meta_data.tables.values()]



import copy
import os.path

import numpy as np

from yangke.common.fileOperate import read_csv_ex, write_lines
from yangke.common.config import logger
from yangke.base import merge_dataframes_simple

folder = r"D:\lengduan"
files = [
    r"origin_retrain_0.csv",
    r"origin_retrain_1.csv",
    r"origin_retrain_2.csv",
    r"origin_retrain_3.csv",
    r"origin_retrain_4.csv",
    r"origin_retrain_5.csv",
    r"origin_retrain_6.csv",
    r"origin_retrain_7.csv",
    r"origin_retrain_8.csv",
    r"origin_retrain_9.csv",
    r"origin_retrain_10.csv",
    r"origin_retrain_11.csv",
]


def collect_data(files, unit_num):
    """
    统计files文件中的所有工况数据，根据温度和功率区间将数据划分至不同的数据集

    :param files: 数据文件名，csv格式文件
    :param unit_num: 机组编号
    :return:
    """

    def collect_t(temperature):
        """
        temperature为长度为12的列表，包括了同温度区间的所有工况数据，每一个元素为一个月的数据。
        本方法将12个月的全部数据合并后，按照功率分别分配到不同的功率范围中，返回五个功率范围中的数据的dataframe。
        五个功率范围分别为：200~250， 250~300， 300~350， 350~400， 400以上，单位为MW

        :param temperature:
        :return:
        """
        p1 = []
        p2 = []
        p3 = []
        p4 = []
        p5 = []
        for x in temperature:
            if x.get("p:200~250") is not None:
                p1.append(copy.deepcopy(x.get("p:200~250")))
            if x.get("p:250~300") is not None:
                p2.append(copy.deepcopy(x.get("p:250~300")))
            if x.get("p:300~350") is not None:
                p3.append(copy.deepcopy(x.get("p:300~350")))
            if x.get("p:350~400") is not None:
                p4.append(copy.deepcopy(x.get("p:350~400")))
            if x.get("p:>400") is not None:
                p5.append(copy.deepcopy(x.get("p:>400")))
        p1 = merge_dataframes_simple(p1)
        p2 = merge_dataframes_simple(p2)
        p3 = merge_dataframes_simple(p3)
        p4 = merge_dataframes_simple(p4)
        p5 = merge_dataframes_simple(p5)
        return p1, p2, p3, p4, p5

    data_whole_section = []
    lines = []
    t_list = []
    p_list = []
    for file in files:
        file = os.path.join(folder, file)
        logger.debug(file)
        df = read_csv_ex(file)
        print(f"{df.shape=}")
        if unit_num == 1:
            df.rename(columns={'电功率1': "电功率", '供热流量1': "供热流量", 'FGH入口燃气压力1': "FGH入口燃气压力",
                               'FGH入口燃气温度1': "FGH入口燃气温度", 'FGH水流量1': "FGH水流量", 'TCA水流量1': "TCA水流量",
                               '过热减温水流量1': "过热减温水流量", '再热减温水流量1': "再热减温水流量", '循泵运行台数1': "循泵运行台数",
                               '机力塔风机运行台数1': "机力塔风机运行台数", '循环效率1': "循环效率", '循环热耗率1': "循环热耗率",
                               '背压1': "背压"}, inplace=True)
        elif unit_num == 2:
            df.rename(columns={'电功率2': "电功率", '供热流量2': "供热流量", 'FGH入口燃气压力2': "FGH入口燃气压力",
                               'FGH入口燃气温度2': "FGH入口燃气温度", 'FGH水流量2': "FGH水流量", 'TCA水流量2': "TCA水流量",
                               '过热减温水流量2': "过热减温水流量", '再热减温水流量2': "再热减温水流量", '循泵运行台数2': "循泵运行台数",
                               '机力塔风机运行台数2': "机力塔风机运行台数", '循环效率2': "循环效率", '循环热耗率2': "循环热耗率",
                               '背压2': "背压"}, inplace=True)

        t_10 = copy.deepcopy(df[df['环境温度'] < 10])
        t_20 = copy.deepcopy(df[(10 <= df['环境温度']) & (df['环境温度'] < 20)])
        t_30 = copy.deepcopy(df[(20 <= df['环境温度']) & (df['环境温度'] < 30)])
        t_40 = copy.deepcopy(df[(30 <= df['环境温度']) & (df['环境温度'] < 40)])
        t_50 = copy.deepcopy(df[40 <= df['环境温度']])
        t_range = {"<10": t_10, "10~20": t_20, "20~30": t_30, "30~40": t_40, ">40": t_50}
        range_statistics = {}
        for k, _t in t_range.items():
            if _t.shape[0] > 0:
                p_10 = _t[_t["电功率"] < 10]
                p_200 = _t[(_t["电功率"] >= 10) & (_t["电功率"] < 200)]
                p_250 = _t[(_t["电功率"] >= 200) & (_t["电功率"] < 250)]
                p_300 = _t[(_t["电功率"] >= 250) & (_t["电功率"] < 300)]
                p_350 = _t[(_t["电功率"] >= 300) & (_t["电功率"] < 350)]
                p_400 = _t[(_t["电功率"] >= 350) & (_t["电功率"] < 400)]
                p_500 = _t[_t["电功率"] >= 400]
            else:
                p_10 = None
                p_200 = None
                p_250 = None
                p_300 = None
                p_350 = None
                p_400 = None
                p_500 = None
            p_t_range = {f"t:{k}": {f"p:停机": p_10, f"p:10~200": p_200,
                                    f"p:200~250": p_250, f"p:250~300": p_300,
                                    f"p:300~350": p_350, f"p:350~400": p_400,
                                    f"p:>400": p_500,
                                    }}
            range_statistics.update(p_t_range)

        cols = 0
        for t, v in range_statistics.items():
            t_list.append(t)
            num_p_list = []
            for p, _ in v.items():
                if p not in p_list:
                    p_list.append(p)
                num = 0 if _ is None else _.shape[0]
                num_p_list.append(f"{num:>8}")

            lines.append(",".join(num_p_list))
            cols = len(num_p_list)

        lines.append("," * (cols - 1))
        data_whole_section.append(copy.deepcopy(range_statistics))
    lines.insert(0, ",".join(p_list))
    write_lines("temp.csv", lines)
    t_10 = [x.get('t:<10') for x in data_whole_section]
    t_20 = [x.get('t:10~20') for x in data_whole_section]
    t_30 = [x.get('t:20~30') for x in data_whole_section]
    t_40 = [x.get('t:30~40') for x in data_whole_section]
    pt_10_250, pt_10_300, pt_10_350, pt_10_400, pt_10_450 = collect_t(t_10)
    pt_20_250, pt_20_300, pt_20_350, pt_20_400, pt_20_450 = collect_t(t_20)
    pt_30_250, pt_30_300, pt_30_350, pt_30_400, pt_30_450 = collect_t(t_30)
    pt_40_250, pt_40_300, pt_40_350, pt_40_400, pt_40_450 = collect_t(t_40)
    return {
        "p:200~250,t:<10": pt_10_250, "p:250~300,t:<10": pt_10_300, "p:300~350,t:<10": pt_10_350,
        "p:350~400,t:<10": pt_10_400, "p:>400,t:<10": pt_10_450,
        "p:200~250,t:10~20": pt_20_250, "p:250~300,t:10~20": pt_20_300, "p:300~350,t:10~20": pt_20_350,
        "p:350~400,t:10~20": pt_20_400, "p:>400,t:10~20": pt_20_450,
        "p:200~250,t:20~30": pt_30_250, "p:250~300,t:20~30": pt_30_300, "p:300~350,t:20~30": pt_30_350,
        "p:350~400,t:20~30": pt_30_400, "p:>400,t:20~30": pt_30_450,
        "p:200~250,t:30~40": pt_40_250, "p:250~300,t:30~40": pt_40_300, "p:300~350,t:30~40": pt_40_350,
        "p:350~400,t:30~40": pt_40_400, "p:>400,t:30~40": pt_40_450,
    }


def dispatch(df_pt_section):
    """
    将df中的数据按照循泵和风机数量划分到不同的数据集中，每个数据集对应一个单独的dataframe。返回的数据及分别表示：
    其中，第一个数据表示循泵运行台数，第二个数字表示风机运行台数。例如pf23表示2泵3风机

    :param df_pt_section:
    :return:
    """

    def dispatch_fan(pump):
        """
        将数据按照风机运行台数划分到[1,2,3,4,5]五个区间中
        返回一个列表，列表的每个元素分别代表风机运行1台、2台、3台、4台、5台时的时间
        :param pump:
        :return:
        """
        p = []
        for i in range(1, 6, 1):
            _ = pump[pump["机力塔风机运行台数"] == i]
            num = _.shape[0]
            power = _["电功率"].mean()  # 统计工况发电量时使用, MW.h
            p.append(power)
            # p.append(num)
        p = np.array(p)
        # 某些工况下风机正在启停，运行台数是错误的
        # if pump.shape[0] > p.sum() > 0:
        #     k_ = pump.shape[0] / p.sum()
        #     p = p * k_
        return p

    p_data1 = df_pt_section[df_pt_section["循泵运行台数"] == 1]
    p_data2 = df_pt_section[df_pt_section["循泵运行台数"] == 2]
    p_data3 = df_pt_section[df_pt_section["循泵运行台数"] == 3]

    p1 = dispatch_fan(p_data1)
    p2 = dispatch_fan(p_data2)
    p3 = dispatch_fan(p_data3)
    # total = p1.sum() + p2.sum() + p3.sum()
    # k_ = df_pt_section.shape[0] / total
    # p1, p2, p3 = p1 * k_, p2 * k_, p3 * k_
    return p1, p2, p3


pt_section = collect_data(files, unit_num=2)
lines = []
for k, v in pt_section.items():
    logger.debug(f"功率和温度范围为：{k}")
    lines.append(f"功率和温度范围为：{k}")
    pump1, pump2, pump3 = dispatch(v)
    logger.info(pump1)
    logger.info(pump2)
    logger.info(pump3)
    lines.append(",".join([str(x) for x in list(pump1)]))
    lines.append(",".join([str(x) for x in list(pump2)]))
    lines.append(",".join([str(x) for x in list(pump3)]))

write_lines(file="temp1.csv", lines=lines)



"""
本程序涉及到的所有物理参数的单位均为国际标准单位
"""
import math
import numpy as np

pressure_larynx_total = 160.70  # 喉部压力全压 Pa
pressure_larynx_dynamic = 160.7 - 120.49  # 喉部压力动压
pressure_larynx_static = 120.49  # 喉部压力静压 Pa
temperature_water_in = 39.06 + 273.15  # K
temperature_water_out = 30.68 + 273.15  # K
temperature_environment_dry = 30.29 + 273.15  # K
temperature_environment_wet = 23.20 + 273.15  # K
pressure_environment = 99470  # Pa
flowrate_water = 5889 / 3600  # m3/s
flowrate_air = 255.97 * 10000 / 3600  # m3/s
ratio_air2water = 0.4844  # 气水比


def get_partial_pressure_of_saturated_vapour(t):
    """
    计算指定温度下饱和水蒸汽的分压

    计算公式参考DL/T1027-2006电力部《工业冷却塔测试规程》第13-14页

    :param t: 温度，K
    :return:
    """
    log_p = 5.005717 - 3.142305 * (1000 / t - 1000 / 373.116) + 8.2 * math.log10(373.16 / t) - 0.0024804 * (373.16 - t)
    p = math.pow(10, log_p)
    return p


def get_relative_humidity_of_moist_air(t_dry, t_wet, p_env=101325, a=3):
    """
    计算空气相对湿度

    参考DL/T1027-2006电力部《工业冷却塔测试规程》第13-14页
    系数A取值参考DL/T1027-2006电力部《工业冷却塔测试规程》第8页

    :param t_dry: 干球温度，K
    :param t_wet: 湿球温度，K
    :param p_env: 环境压力，Pa
    :param a: 干湿球温度计系数列表，可以取0~4，分别对应阿斯曼通风干湿表，标准百叶箱通风干湿表，阿费古斯特湿度表，百叶箱柱状干湿表，百叶箱球状干湿表
    :return: 相对湿度fai，取值范围为0~1
    """
    a_list = [0.000662, 0.000667, 0.0007947, 0.000815, 0.000857]
    p_wet = get_partial_pressure_of_saturated_vapour(t_wet)
    p_dry = get_partial_pressure_of_saturated_vapour(t_dry)
    return (p_wet - a_list[a] * p_env * (t_dry - t_wet)) / p_dry


def get_rho_of_moist_air(fai, t, p_env=101325):
    """
    计算湿空气密度

    :param fai: 空气的相对湿度，%，即空气中水汽压与饱和水汽压的百分比，计算取值范围为0~1
    :param t: 空气的开尔文温度
    :param p_env: 大气压力，Pa，默认101325
    :return:
    """
    return 1 / t * (0.003483 * p_env - 0.001316 * fai * get_partial_pressure_of_saturated_vapour(t))


def get_moisture_of_moist_air(fai, t, p_env=101325):
    """
    计算空气的含湿量，即湿空气中与一千克干空气同时并存的水蒸气的质量（克）

    :param fai: 空气的相对湿度，%，即空气中水汽压与饱和水汽压的百分比
    :param t: 温度，K
    :param p_env: 大气压力，Pa，默认101325
    :return:
    """
    p_theta = get_partial_pressure_of_saturated_vapour(t)  # 饱和空气中水蒸汽的分压
    return 0.622 * fai * p_theta / (p_env - fai * p_theta)


def get_enthalpy_of_moist_air(t_dry, moisture, gama=2500000, cd=1005, cv=1846):
    """
    计算湿空气的比焓

    :param t_dry: 空气的干球温度
    :param moisture: 空气的含湿量
    :param gama: 水在0℃时的汽化热
    :param cd: 干空气的比热容
    :param cv: 水蒸汽的比热容
    :return: 返回比焓，注意单位为J/kg
    """
    return cd * t_dry + moisture * (gama + cv * t_dry)


def get_air_flow_m3ph(f, rho2, p, k0=1):
    """
    冷却塔进塔空气量计算

    :param f: 有效通风面积
    :param k0: 毕托管校正系数，一般为1
    :param rho2: 出塔空气密度
    :param p: 风筒喉部动压
    :return:
    """
    return f * k0 * math.sqrt(2 / rho2) * p


def cal_air_flow_p_from_exp(p_d):
    """
    根据测量值计算风机喉部动压

    一般在风筒喉部四个方向打小孔，深入皮托管测量动压值，在喉部截面不同半径处测量得到一系列动压值，本方法根据所有测得的动压值计算平均动压

    :param p_d: 测点的动压值列表
    :return:
    """
    return np.array(p_d).mean()


def fan_pressure_total(rho_a, v1_m, v2_m, delta_p, zita=0.16):
    """
    计算风机全压，风机全压即整塔阻力

    塔进风口至风机下的总阻力为delta_p，为风量测量截面各测点全压与大气压插值的算数平均值

    :param rho_a: 出塔空气密度
    :param v1_m: 风量测量断面平均风速
    :param v2_m: 风筒出口断面平均风速
    :param delta_p: 塔进风口至风机下的总阻力，为风量测量断面各测点全压与大气压差值的算数平均值
    :param zita: 风筒气流阻力系数。与风筒扩散角有关，默认取值0.16
    :return:
    """
    ksi = zita * (1 - math.pow(v2_m / v1_m, 2))
    delta_p_f = 0.5 * rho_a * v2_m * v2_m + ksi * 0.5 * rho_a * v1_m * v1_m - delta_p
    return delta_p_f


def cal_omega(t1, t2, n=20, c_w=4.1868, method='simpson'):
    """
    计算冷却塔冷却数

    逆流塔热力数学模型是 Ka*V/Q=integrate(c_w/(h_t-h_theta), (t, t2, t1))，冷却塔数及方程右侧的积分数值，参考冷却塔性能试验规程
    式中，h_t是与水温相应的饱和空气比焓，h_theta是湿空气比焓，按进塔空气湿度计算

    :param t1: 冷却塔进水温度
    :param t2: 冷却塔出水温度
    :param n: simpson积分的分段数，最好取偶数
    :param c_w: 4.1868kJ/(kgK)，水的比热
    :param method: 积分方法，‘simpson'，’Chebyshev‘
    :return:
    """
    if method != 'simpson':
        print('暂不支持其他积分方法')
        return None
    delta_t = t1 - t2
    t_list = []  # 将t2到t1的温度等分为n段，分段的温度列表，对应t2, t2+dt, t2+2dt, t2+3dt,...
    h_list = []  # 对应于t_list中温度的饱和空气的比焓
    h_theta_list = []  # 第i个等份的空气焓
    dt = delta_t / (n - 1)

    for i in range(n):
        t_list.append(t2 + dt * i)
        # todo
        h_list.append()
        k_i = 1 - (t2 + i * dt) / (586 - 0.56 * (t2 + i * dt - 20))
        h_theta_list.append()
    sigma = 0
    for i in range(n):
        if i == 0 or i == (n - 1):
            sigma = sigma + 1 / (h_list[i] - h_theta_list[i])
        elif i % 2 == 0:
            sigma = sigma + 2 / (h_list[i] - h_theta_list[i])
        else:
            sigma = sigma + 4 / (h_list[i] - h_theta_list[i])
    return c_w * delta_t * sigma / 3 / n


